# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# Copyright (c) 2023 Brian Pugh
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import abc
from dataclasses import dataclass
from time import time
from binascii import crc32
from typing import (Any, List, Optional, Union)

from ..core.target import Target
from ..core.exceptions import (FlashFailure, FlashProgramFailure)
from ..core.memory_map import MemoryRegion
from ..utility.mask import same

# Number of bytes in a page to read to quickly determine if the page has the same data
PAGE_ESTIMATE_SIZE = 32
DATA_TRANSFER_B_PER_S = 40 * 1000 # ~40KB/s, depends on clock speed, theoretical limit for HID is 56,000 B/s

LOG = logging.getLogger(__name__)

def get_page_count(count: int) -> str:
    """@brief Return string for page count with correct plurality."""
    if count == 1:
        return "1 page"
    else:
        return "{} pages".format(count)

def get_sector_count(count: int) -> str:
    """@brief Return string for sector count with correct plurality."""
    if count == 1:
        return "1 sector"
    else:
        return "{} sectors".format(count)

@dataclass
class ProgrammingInfo:
    program_type: Any = None                # Type of programming performed - FLASH_SECTOR_ERASE or FLASH_CHIP_ERASE
    program_time: float = 0.0               # Total programming time
    analyze_type: Any = None                # Type of flash analysis performed - FLASH_ANALYSIS_CRC32 or FLASH_ANALYSIS_PARTIAL_PAGE_READ
    analyze_time: float = 0.0               # Time to analyze flash contents
    total_byte_count: int = 0
    program_byte_count: int = 0
    program_page_count: int = 0
    erase_byte_count: int = 0
    erase_sector_count: int = 0
    skipped_byte_count: int = 0
    skipped_page_count: int = 0

class MemoryBuilder(abc.ABC):
    """@brief Abstract class for memory builders."""

    def __init__(self) -> None:
        super().__init__()
        self._buffered_data_size: int = 0

    @property
    def buffered_data_size(self) -> int:
        """@brief Total amount of memory buffered by this builder."""
        return self._buffered_data_size

    @property
    @abc.abstractmethod
    def region(self) -> MemoryRegion:
        """@brief The memory region that can be written to by this builder."""
        ...

    @abc.abstractmethod
    def add_data(self, addr: int, data: Union[bytes, bytearray]) -> None:
        """@brief Add a chunk of data to the builder."""
        ...

    @abc.abstractmethod
    def program(self, **kwargs: Any) -> ProgrammingInfo:
        """@brief Commit the buffered data to the destination memory region."""
        ...

def _stub_progress(percent):
    pass

class _FlashSector:
    """@brief Info about an erase sector and all pages to be programmed within it."""
    def __init__(self, sector_info, n_subsectors: int = 1):
        self.addr: int = sector_info.base_addr
        self._subsector_size: int = sector_info.size
        self.max_page_count: int = 0
        self.page_list: List[_FlashPage] = []
        self.erase_weight: float = sector_info.erase_weight
        self.n_subsectors = n_subsectors

    @property
    def size(self):
        return self.n_subsectors * self._subsector_size

    @property
    def addrs(self):
        for i in range(self.n_subsectors):
            yield self.addr + (i * self._subsector_size)

    def add_page(self, page):
        # The first time a page is added, compute the page count for this sector. This
        # obviously assumes that all the pages in the sector are the same size.
        if len(self.page_list) == 0:
            self.max_page_count = self.size // page.size
            assert (self.size % page.size) == 0, "Flash pages (%d bytes) do not fit evenly " \
                                                "into sector (%d bytes)" % (page.size, self.size)
        assert len(self.page_list) < self.max_page_count
        self.page_list.append(page)
        self.page_list.sort(key=lambda p:p.addr)

    def are_any_pages_not_same(self):
        """@brief Returns True if any pages in this sector might need to be programmed."""
        return any(page.same is not True for page in self.page_list)

    def mark_all_pages_not_same(self):
        """@brief Sets the same flag to False for all pages in this sector."""
        for page in self.page_list:
            page.same = False

    def __repr__(self):
        return "<_FlashSector@%x addr=%x size=%x wgt=%g pages=%s, subsectors=%d>" % (
            id(self), self.addr, self.size, self.erase_weight, self.page_list, self.n_subsectors)

class _FlashPage:
    """@brief A page to be programmed and its data."""
    def __init__(self, page_info):
        self.addr: int = page_info.base_addr
        self.size: int = page_info.size
        self.data: List[int] = []
        self.program_weight: float = page_info.program_weight
        self.erased: Optional[bool] = None # Whether the data all matches the erased value.
        self.same: Optional[bool] = None
        self.crc: int = 0
        self.cached_estimate_data: Optional[List[int]] = None

    def get_program_weight(self):
        """@brief Get time to program a page including the data transfer."""
        return self.program_weight + \
            float(len(self.data)) / float(DATA_TRANSFER_B_PER_S)

    def get_verify_weight(self):
        """@brief Get time to verify a page."""
        return float(self.size) / float(DATA_TRANSFER_B_PER_S)

    def __repr__(self):
        return "<_FlashPage@%x addr=%x size=%x datalen=%x wgt=%g erased=%s same=%s>" % (
            id(self), self.addr, self.size, len(self.data), self.program_weight, self.erased, self.same)

class _FlashOperation:
    """@brief Holds requested data to be programmed at a given address."""
    def __init__(self, addr, data):
        self.addr = addr
        self.data = data

class FlashBuilder(MemoryBuilder):
    """@brief Manages programming flash within one flash memory region.

    The purpose of this class is to optimize flash programming within a single region to achieve
    the highest flash programming performance possible. Various methods are used to estimate the
    fastest programming method.

    Individual flash algorithm operations are performed by the @ref pyocd.flash.flash.Flash
    "Flash" instance provided to the contructor.

    Assumptions:
    1. Sector erases must be on sector boundaries.
    2. Page writes must be on page boundaries.
    3. Pages are never larger than sectors, but can be smaller.
    4. There must be an even number of pages within a sector.
    5. Entire pages must be programmed.
    """

    # Type of flash operation
    FLASH_SECTOR_ERASE = 1
    FLASH_CHIP_ERASE = 2

    # Type of flash analysis
    FLASH_ANALYSIS_CRC32 = "CRC32"
    FLASH_ANALYSIS_PARTIAL_PAGE_READ = "PAGE_READ"

    def __init__(self, flash):
        super().__init__()
        self.flash = flash
        self.flash_start = flash.region.start
        self.flash_operation_list = []
        self.sector_list: List[_FlashSector] = []
        self.page_list: List[_FlashPage] = []
        self.perf = ProgrammingInfo()
        self.enable_double_buffering = True
        self.log_performance = True
        self._buffered_data_size = 0
        self.program_byte_count = 0
        self.sector_erase_count = 0
        self.chip_erase_count = 0 # Number of pages to program using chip erase method.
        self.chip_erase_weight = 0 # Erase/program weight using chip erase method.
        self.sector_erase_count = 0 # Number of pages to program using sector erase method.
        self.sector_erase_weight = 0 # Erase/program weight using sector erase method.
        self.algo_inited_for_read = False

    @property
    def region(self) -> MemoryRegion:
        return self.flash.region

    def enable_double_buffer(self, enable):
        self.enable_double_buffering = enable

    def add_data(self, addr, data):
        """@brief Add a block of data to be programmed.

        @note Programming does not start until the method program() is called.

        @param self
        @param addr Base address of the block of data passed to this method. The entire block of
            data must be contained within the flash memory region associated with this instance.
        @param data Data to be programmed. Should be a list of byte values.

        @exception ValueError Attempt to add overlapping data, or address range of added data is
            outside the address range of the flash region associated with the builder.
        """
        # Ignore empty data.
        if len(data) == 0:
            return

        # Sanity check
        if not self.flash.region.contains_range(start=addr, length=len(data)):
            raise ValueError("Flash address range 0x%x-0x%x is not contained within region '%s'" %
                (addr, addr + len(data) - 1, self.flash.region.name))

        # Add operation to list
        self.flash_operation_list.append(_FlashOperation(addr, data))
        self._buffered_data_size += len(data)

        # Keep list sorted
        self.flash_operation_list = sorted(self.flash_operation_list, key=lambda operation: operation.addr)

        # Verify this does not overlap
        prev_flash_operation = None
        for operation in self.flash_operation_list:
            if prev_flash_operation is not None:
                if prev_flash_operation.addr + len(prev_flash_operation.data) > operation.addr:
                    raise ValueError("Error adding data - Data at 0x%x..0x%x overlaps with 0x%x..0x%x"
                            % (prev_flash_operation.addr, prev_flash_operation.addr + len(prev_flash_operation.data),
                               operation.addr, operation.addr + len(operation.data)))
            prev_flash_operation = operation

    def _enable_read_access(self):
        """@brief Ensure flash is accessible by initing the algo for verify.

        Not all flash memories are always accessible. For instance, external QSPI. Initing the
        flash algo for the VERIFY operation is the canonical way to ensure that the flash is
        memory mapped and accessible.
        """
        if not self.algo_inited_for_read:
            try:
                self.flash.init(self.flash.Operation.VERIFY)
            except FlashFailure:
                # If initing for verify fails, then try again in erase mode.
                self.flash.init(self.flash.Operation.ERASE)
            self.algo_inited_for_read = True

    def _build_sectors_and_pages(self, keep_unwritten):
        """@brief Converts the list of flash operations to flash sectors and pages.

        @param self
        @param keep_unwritten If true, unwritten pages in an erased sector and unwritten
            contents of a modified page will be read from the target and added to the data to be
            programmed.

        @exception FlashFailure Could not get sector or page info for an address.
        """
        assert len(self.flash_operation_list) > 0

        self.program_byte_count = 0

        flash_addr = self.flash_operation_list[0].addr
        sector_info = self.flash.get_sector_info(flash_addr)
        if sector_info is None:
            raise FlashFailure("attempt to program invalid flash address", address=flash_addr)

        page_info = self.flash.get_page_info(flash_addr)
        if page_info is None:
            raise FlashFailure("attempt to program invalid flash address", address=flash_addr)

        def create_flash_sector(sector_info, page_info):
            if page_info.size > sector_info.size:
                assert page_info.size % sector_info.size == 0, \
                    f"Sector ({sector_info.size} bytes) do not fit evenly into page ({page_info.size} bytes)"
                n_subsectors = page_info.size // sector_info.size
            else:
                n_subsectors = 1

            return _FlashSector(sector_info, n_subsectors=n_subsectors)

        current_page = _FlashPage(page_info)
        current_sector = create_flash_sector(sector_info, page_info)

        self.page_list.append(current_page)
        self.sector_list.append(current_sector)

        current_sector.add_page(current_page)

        def fill_end_of_page_gap():
            # Fill the gap at the end of the soon to be previous page if there is one
            if len(current_page.data) != current_page.size:
                page_data_end = current_page.addr + len(current_page.data)
                old_data_len = current_page.size - len(current_page.data)
                if keep_unwritten and self.flash.region.is_readable:
                    self._enable_read_access()
                    old_data = self.flash.target.read_memory_block8(page_data_end, old_data_len)
                else:
                    old_data = [self.flash.region.erased_byte_value] * old_data_len
                current_page.data.extend(old_data)
                self.program_byte_count += old_data_len

        for flash_operation in self.flash_operation_list:
            pos = 0
            while pos < len(flash_operation.data):
                flash_addr = flash_operation.addr + pos

                # Check if operation is in a different sector.
                if flash_addr >= current_sector.addr + current_sector.size:
                    sector_info = self.flash.get_sector_info(flash_addr)
                    if sector_info is None:
                        raise FlashFailure("attempt to program invalid flash address", address=flash_addr)
                    current_sector = create_flash_sector(sector_info, page_info)
                    self.sector_list.append(current_sector)

                # Check if operation is in a different page
                if flash_addr >= current_page.addr + current_page.size:
                    # Fill any gap at the end of the current page before switching to a new page.
                    fill_end_of_page_gap()

                    # Create the new page.
                    page_info = self.flash.get_page_info(flash_addr)
                    if page_info is None:
                        raise FlashFailure("attempt to program invalid flash address", address=flash_addr)
                    current_page = _FlashPage(page_info)
                    current_sector.add_page(current_page)
                    self.page_list.append(current_page)

                # Fill the page gap if there is one
                page_data_end = current_page.addr + len(current_page.data)
                if flash_addr != page_data_end:
                    old_data_len = flash_addr - page_data_end
                    if keep_unwritten and self.flash.region.is_readable:
                        self._enable_read_access()
                        old_data = self.flash.target.read_memory_block8(page_data_end, old_data_len)
                    else:
                        old_data = [self.flash.region.erased_byte_value] * old_data_len
                    current_page.data.extend(old_data)
                    self.program_byte_count += old_data_len

                # Copy data to page and increment pos
                space_left_in_page = page_info.size - len(current_page.data)
                space_left_in_data = len(flash_operation.data) - pos
                amount = min(space_left_in_page, space_left_in_data)
                current_page.data.extend(flash_operation.data[pos:pos + amount])
                self.program_byte_count += amount

                #increment position
                pos += amount

        # Fill the page gap at the end if there is one
        fill_end_of_page_gap()

        # Go back through sectors and fill any missing pages with existing data.
        if keep_unwritten and self.flash.region.is_readable:
            self._fill_unwritten_sector_pages()

    def _fill_unwritten_sector_pages(self):
        """@brief Fill in missing pages from sectors we are going to modify."""
        for sector in self.sector_list:
            sector_page_number = 0
            sector_page_addr = sector.addr

            def add_page_with_existing_data():
                page_info = self.flash.get_page_info(sector_page_addr)
                if page_info is None:
                    raise FlashFailure("attempt to program invalid flash address", address=sector_page_addr)
                new_page = _FlashPage(page_info)
                self._enable_read_access()
                new_page.data = self.flash.target.read_memory_block8(new_page.addr, new_page.size)
                new_page.same = True
                sector.add_page(new_page)
                self.page_list.append(new_page)
                self.program_byte_count += len(new_page.data)
                return new_page

            # Iterate over pages defined for the sector. If a gap is found, a new page is inserted
            # with the current contents of target memory.
            while sector_page_number < len(sector.page_list):
                page = sector.page_list[sector_page_number]

                if page.addr != sector_page_addr:
                    page = add_page_with_existing_data()

                sector_page_number += 1
                sector_page_addr += page.size

            # Add missing pages at the end of the sector.
            while sector_page_addr < sector.addr + sector.size:
                page = add_page_with_existing_data()
                sector_page_addr += page.size

    def program(self, chip_erase=None, progress_cb=None, smart_flash=True, fast_verify=False, keep_unwritten=True, no_reset=False):
        """@brief Determine fastest method of flashing and then run flash programming.

        Data must have already been added with add_data().

        If the flash region's 'are_erased_sectors_readable' attribute is false, then the
        smart_flash, fast_verify, and keep_unwritten options are forced disabled.

        @param self
        @param chip_erase A value of "chip" forces chip erase, "sector" forces sector erase, and a
            value of "auto" means that the estimated fastest method should be used. If not
            specified, the default is auto.
        @param progress_cb A callable that accepts a single parameter of the percentage complete.
        @param smart_flash If True, FlashBuilder will scan target memory to attempt to avoid
            programming contents that are not changing with this program request. False forces
            all requested data to be programmed.
        @param fast_verify If smart_flash is enabled and the target supports the CRC32 analyzer,
            this parameter controls whether positive results from the analyzer will be accepted.
            In other words, pages with matching CRCs will be marked as the same. There is a small,
            but non-zero, chance that the CRCs match even though the data is different, but the
            odds of this happing are low: ~1/(2^32) = ~2.33*10^-8%.
        @param keep_unwritten Depending on the sector versus page size and the amount of data
            written, there may be ranges of flash that would be erased but not written with new
            data. This parameter sets whether the existing contents of those unwritten ranges will
            be read from memory and restored while programming.
        @param no_reset Boolean indicating whether if the device should not be reset after the
            programming process has finished.
        """

        # Send notification that we're about to program flash.
        self.flash.target.session.notify(Target.Event.PRE_FLASH_PROGRAM, self)

        # Disable options if attempting to read erased sectors will fault.
        if not self.flash.region.are_erased_sectors_readable:
            smart_flash = False
            fast_verify = False
            keep_unwritten = False

        # Examples
        # - lpc4330     -Non 0 base address
        # - nRF51       -UICR location far from flash (address 0x10001000)
        # - LPC1768     -Different sized pages
        program_start = time()

        if progress_cb is None:
            progress_cb = _stub_progress

        # There must be at least 1 flash operation
        if len(self.flash_operation_list) == 0:
            LOG.warning("No pages were programmed")
            return

        # Convert chip_erase.
        if (chip_erase is None) or (chip_erase == "auto"):
            chip_erase = None
        elif chip_erase == "sector":
            chip_erase = False
        elif chip_erase == "chip":
            chip_erase = True
        else:
            raise ValueError("invalid chip_erase value '{}'".format(chip_erase))

        # Convert the list of flash operations into flash sectors and pages
        self._build_sectors_and_pages(keep_unwritten)
        assert len(self.sector_list) != 0 and len(self.sector_list[0].page_list) != 0
        self.flash_operation_list = [] # Don't need this data in memory anymore.

        # If smart flash was set to false then mark all pages
        # as requiring programming
        if not smart_flash:
            self._mark_all_pages_for_programming()

        # If the flash algo doesn't support erase all, disable chip erase.
        if not self.flash.is_erase_all_supported:
            chip_erase = False

        # If the first page being programmed is not the first page
        # in flash then don't use a chip erase unless explicitly directed to.
        if self.page_list[0].addr > self.flash_start:
            if chip_erase is None:
                chip_erase = False
            elif chip_erase is True:
                LOG.warning('Chip erase used when flash address 0x%x is not the same as flash start 0x%x',
                    self.page_list[0].addr, self.flash_start)

        chip_erase_count, chip_erase_program_time = self._compute_chip_erase_pages_and_weight()
        sector_erase_min_program_time = self._compute_sector_erase_pages_weight_min()

        # If chip_erase hasn't been specified determine if chip erase is faster
        # than page erase regardless of contents
        if (chip_erase is None) and (chip_erase_program_time < sector_erase_min_program_time):
            chip_erase = True

        # If chip erase isn't True then analyze the flash
        if chip_erase is not True:
            sector_erase_count, page_program_time = self._compute_sector_erase_pages_and_weight(fast_verify)
        else:
            sector_erase_count, page_program_time = 0, 0

        # If chip erase hasn't been set then determine fastest method to program
        if chip_erase is None:
            LOG.debug("Chip erase count %i, sector erase est count %i" % (chip_erase_count, sector_erase_count))
            LOG.debug("Chip erase weight %f, sector erase weight %f" % (chip_erase_program_time, page_program_time))
            chip_erase = chip_erase_program_time < page_program_time

        if chip_erase:
            if self.flash.is_double_buffering_supported and self.enable_double_buffering:
                LOG.debug("Using double buffer chip erase program")
                flash_operation = self._chip_erase_program_double_buffer(progress_cb)
            else:
                flash_operation = self._chip_erase_program(progress_cb)
        else:
            if self.flash.is_double_buffering_supported and self.enable_double_buffering:
                LOG.debug("Using double buffer sector erase program")
                flash_operation = self._sector_erase_program_double_buffer(progress_cb)
            else:
                flash_operation = self._sector_erase_program(progress_cb)

        # Cleanup flash algo and reset target after programming.
        self.flash.cleanup()

        if no_reset is not True:
            self.flash.target.reset_and_halt()

        program_finish = time()
        self.perf.program_time = program_finish - program_start
        self.perf.program_type = flash_operation

        erase_byte_count = 0
        erase_sector_count = 0
        actual_program_byte_count = 0
        actual_program_page_count = 0
        skipped_byte_count = 0
        skipped_page_count = 0
        for page in self.page_list:
            if (page.same is True) or (page.erased and chip_erase):
                skipped_byte_count += page.size
                skipped_page_count += 1
            else:
                actual_program_byte_count += page.size
                actual_program_page_count += 1
        for sector in self.sector_list:
            if sector.are_any_pages_not_same():
                erase_byte_count += sector.size
                erase_sector_count += 1

        self.perf.total_byte_count = self.program_byte_count
        self.perf.program_byte_count = actual_program_byte_count
        self.perf.program_page_count = actual_program_page_count
        self.perf.erase_byte_count = erase_byte_count
        self.perf.erase_sector_count = erase_sector_count
        self.perf.skipped_byte_count = skipped_byte_count
        self.perf.skipped_page_count = skipped_page_count

        if self.log_performance:
            if chip_erase:
                LOG.info("Erased chip, programmed %d bytes (%s), skipped %d bytes (%s) at %.02f kB/s",
                    actual_program_byte_count, get_page_count(actual_program_page_count),
                    skipped_byte_count, get_page_count(skipped_page_count),
                    ((self.program_byte_count/1024) / self.perf.program_time))
            else:
                LOG.info("Erased %d bytes (%s), programmed %d bytes (%s), skipped %d bytes (%s) at %.02f kB/s",
                    erase_byte_count, get_sector_count(erase_sector_count),
                    actual_program_byte_count, get_page_count(actual_program_page_count),
                    skipped_byte_count, get_page_count(skipped_page_count),
                    ((self.program_byte_count/1024) / self.perf.program_time))

        # Send notification that we're done programming flash.
        self.flash.target.session.notify(Target.Event.POST_FLASH_PROGRAM, self)

        return self.perf

    def get_performance(self):
        return self.perf

    def _mark_all_pages_for_programming(self):
        for sector in self.sector_list:
            sector.erased = False
            for page in sector.page_list:
                sector.erased = False
                page.same = False

    def _compute_chip_erase_pages_and_weight(self):
        """@brief Compute the number of erased pages.

        Determine how many pages in the new data are already erased.
        """
        chip_erase_count = 0
        chip_erase_weight = 0
        chip_erase_weight += self.flash.get_flash_info().erase_weight
        for page in self.page_list:
            if page.erased is None:
                page.erased = self.flash.region.is_data_erased(page.data)
            if not page.erased:
                chip_erase_count += 1
                chip_erase_weight += page.get_program_weight()
        self.chip_erase_count = chip_erase_count
        self.chip_erase_weight = chip_erase_weight
        return chip_erase_count, chip_erase_weight

    def _compute_sector_erase_pages_weight_min(self):
        return sum(page.get_verify_weight() for page in self.page_list)

    def _analyze_pages_with_partial_read(self):
        """@brief Estimate how many pages are the same by reading data.

        Pages are analyzed by reading the first 32 bytes and comparing with data to be
        programmed.
        """
        # Quickly estimate how many pages are the same as current flash contents.
        # Init the flash algo in case it is required in order to access the flash memory.
        self._enable_read_access()
        for page in self.page_list:
            # Analyze pages that haven't been analyzed yet
            if page.same is None:
                size = min(PAGE_ESTIMATE_SIZE, len(page.data))
                data = self.flash.target.read_memory_block8(page.addr, size)
                page_same = same(data, page.data[0:size])
                if page_same is False:
                    page.same = False
                else:
                    # Save the data read for estimation so we don't need to read it again.
                    page.cached_estimate_data = data

    def _analyze_pages_with_crc32(self, assume_estimate_correct=False):
        """@brief Estimate how many pages are the same using a CRC32 analyzer.

        A CRC32 analyzer program is loaded into target RAM and is passed an array of pages
        and sizes. When executed, it computes the CRC32 for every page.

        @param self
        @param assume_estimate_correct If set to True, then pages with matching CRCs will
            be marked as the same.  There is a small chance that the CRCs match even though the
            data is different, but the odds of this happing are low: ~1/(2^32) = ~2.33*10^-8%.
        """
        # Build list of all the pages that need to be analyzed
        sector_list = []
        page_list = []
        for page in self.page_list:
            if page.same is None:
                # Add page to compute_crcs
                sector_list.append((page.addr, page.size))
                page_list.append(page)
                # Compute CRC of data (Padded with 0xFF)
                data = list(page.data)
                pad_size = page.size - len(page.data)
                if pad_size > 0:
                    data.extend([0xFF] * pad_size)
                page.crc = crc32(bytearray(data)) & 0xFFFFFFFF

        # Analyze pages
        if len(page_list) > 0:
            self._enable_read_access()
            crc_list = self.flash.compute_crcs(sector_list)
            for page, crc in zip(page_list, crc_list):
                page_same = page.crc == crc
                if assume_estimate_correct:
                    page.same = page_same
                elif page_same is False:
                    page.same = False

    def _compute_sector_erase_pages_and_weight(self, fast_verify):
        """@brief Quickly analyze flash contents and compute weights for sector erase.

        Quickly estimate how many pages are the same.  These estimates are used
        by _sector_erase_program so it is recommended to call this before beginning programming
        This is done automatically by smart_program.
        """
        analyze_start = time()

        # Analyze unknown pages using either CRC32 analyzer or partial reads.
        if any(page.same is None for page in self.page_list):
            if self.flash.get_flash_info().crc_supported:
                self._analyze_pages_with_crc32(fast_verify)
                self.perf.analyze_type = FlashBuilder.FLASH_ANALYSIS_CRC32
            elif self.flash.region.is_readable:
                self._analyze_pages_with_partial_read()
                self.perf.analyze_type = FlashBuilder.FLASH_ANALYSIS_PARTIAL_PAGE_READ
            else:
                # The CRC analyzer isn't supported and flash isn't directly readable, so
                # just mark all pages as needing programming. This will also prevent
                # _scan_pages_for_same() from trying to read flash.
                self._mark_all_pages_for_programming()

        # Put together page and time estimate.
        sector_erase_count = 0
        sector_erase_weight = 0
        for sector in self.sector_list:
            for page in sector.page_list:
                if page.same is False:
                    sector_erase_count += 1
                    sector_erase_weight += page.get_program_weight()
                elif page.same is None:
                    # Page may be the same but must be read to confirm
                    sector_erase_weight += page.get_verify_weight()
                elif page.same is True:
                    # Page is confirmed to be the same so no programming weight
                    pass

            if sector.are_any_pages_not_same():
                sector_erase_weight += sector.erase_weight

        self.sector_erase_count = sector_erase_count
        self.sector_erase_weight = sector_erase_weight

        analyze_finish = time()
        self.perf.analyze_time = analyze_finish - analyze_start
        LOG.debug("Analyze time: %f" % (analyze_finish - analyze_start))

        return sector_erase_count, sector_erase_weight

    def _chip_erase_program(self, progress_cb=_stub_progress):
        """@brief Program by first performing an erase all."""
        LOG.debug("%i of %i pages have erased data", len(self.page_list) - self.chip_erase_count, len(self.page_list))
        progress_cb(0.0)
        progress = 0

        self.flash.init(self.flash.Operation.ERASE)
        self.flash.erase_all()
        self.flash.uninit()

        progress += self.flash.get_flash_info().erase_weight
        progress_cb(float(progress) / float(self.chip_erase_weight))

        self.flash.init(self.flash.Operation.PROGRAM)
        for page in self.page_list:
            if not page.erased:
                self.flash.program_page(page.addr, page.data)
                progress += page.get_program_weight()
                progress_cb(float(progress) / float(self.chip_erase_weight))
        self.flash.uninit()
        progress_cb(1.0)
        return FlashBuilder.FLASH_CHIP_ERASE

    def _next_unerased_page(self, i):
        if i >= len(self.page_list):
            return None, i
        page = self.page_list[i]
        while page.erased:
            i += 1
            if i >= len(self.page_list):
                return None, i
            page = self.page_list[i]
        return page, i + 1

    def _chip_erase_program_double_buffer(self, progress_cb=_stub_progress):
        """@brief Double-buffered program by first performing an erase all."""
        LOG.debug("%i of %i pages have erased data", len(self.page_list) - self.chip_erase_count, len(self.page_list))
        progress_cb(0.0)
        progress = 0

        program_timeout = self.flash.target.session.options.get('flash.timeout.program')

        self.flash.init(self.flash.Operation.ERASE)
        self.flash.erase_all()
        self.flash.uninit()

        progress += self.flash.get_flash_info().erase_weight
        progress_cb(float(progress) / float(self.chip_erase_weight))

        # Set up page and buffer info.
        current_buf = 0
        next_buf = 1
        page, i = self._next_unerased_page(0)
        assert page is not None

        # Load first page buffer
        self.flash.load_page_buffer(current_buf, page.addr, page.data)

        self.flash.init(self.flash.Operation.PROGRAM)
        while page is not None:
            # Kick off this page program.
            current_addr = page.addr
            current_weight = page.get_program_weight()
            self.flash.start_program_page_with_buffer(current_buf, current_addr)

            # Get next page and load it.
            page, i = self._next_unerased_page(i)
            if page is not None:
                self.flash.load_page_buffer(next_buf, page.addr, page.data)

            # Wait for the program to complete.
            result = self.flash.wait_for_completion(timeout=program_timeout)
            if result == self.flash.TIMEOUT_ERROR:
                raise FlashProgramFailure('flash program page timeout', address=current_addr, result_code=result)
            elif result != 0:
                raise FlashProgramFailure('flash program page failure', address=current_addr, result_code=result)

            # Swap buffers.
            current_buf, next_buf = next_buf, current_buf

            # Update progress.
            progress += current_weight
            progress_cb(float(progress) / float(self.chip_erase_weight))

        self.flash.uninit()
        progress_cb(1.0)
        return FlashBuilder.FLASH_CHIP_ERASE

    def _sector_erase_program(self, progress_cb=_stub_progress):
        """@brief Program by performing sector erases."""
        actual_sector_erase_count = 0
        actual_sector_erase_weight = 0
        progress = 0

        progress_cb(0.0)

        # Fill in same flag for all pages. This is done up front so we're not trying
        # to read from flash while simultaneously programming it.
        progress = self._scan_pages_for_same(progress_cb)

        for sector in self.sector_list:
            if sector.are_any_pages_not_same():
                if self.region.is_erasable:
                    # Erase the sector
                    self.flash.init(self.flash.Operation.ERASE)
                    for addr in sector.addrs:
                        self.flash.erase_sector(addr)
                    self.flash.uninit()

                    actual_sector_erase_weight += sector.erase_weight

                # Update progress
                if self.sector_erase_weight > 0:
                    progress_cb(float(progress) / float(self.sector_erase_weight))

                # The sector was erased, so we must program all pages in the sector
                # regardless of whether they were the same or not.
                for page in sector.page_list:

                    progress += page.get_program_weight()

                    self.flash.init(self.flash.Operation.PROGRAM)
                    self.flash.program_page(page.addr, page.data)
                    self.flash.uninit()

                    actual_sector_erase_count += 1
                    actual_sector_erase_weight += page.get_program_weight()

                    # Update progress
                    if self.sector_erase_weight > 0:
                        progress_cb(float(progress) / float(self.sector_erase_weight))

        progress_cb(1.0)

        LOG.debug("Estimated sector erase programmed page count: %i", self.sector_erase_count)
        LOG.debug("Actual sector erase programmed page count: %i", actual_sector_erase_count)

        return FlashBuilder.FLASH_SECTOR_ERASE

    def _scan_pages_for_same(self, progress_cb=_stub_progress):
        """@brief Read the full page data to determine if it is unchanged.

        When this function exits, the same flag will be set to either True or False for
        every page. In addition, sectors that need at least one page programmed will have
        the same flag set to False for all pages within that sector.
        """
        progress = 0

        # Read page data if unknown - after this page.same will be True or False
        unknown_pages = [page for page in self.page_list if page.same is None]
        if unknown_pages:
            self._enable_read_access()

            for page in unknown_pages:
                if page.cached_estimate_data is not None:
                    data = page.cached_estimate_data
                    offset = len(data)
                else:
                    data = []
                    offset = 0
                assert len(page.data) == page.size, "page data size (%d) != page size (%d)" % (len(page.data), page.size)
                data.extend(self.flash.target.read_memory_block8(page.addr + offset,
                                                                    page.size - offset))
                page.same = same(page.data, data)
                page.cached_estimate_data = None # This data isn't needed anymore.
                progress += page.get_verify_weight()

                # Update progress
                if self.sector_erase_weight > 0:
                    progress_cb(float(progress) / float(self.sector_erase_weight))

        # If we have to program any pages of a sector, then mark all pages of that sector
        # as needing to be programmed, since the sector will be erased.
        for sector in self.sector_list:
            if sector.are_any_pages_not_same():
                sector.mark_all_pages_not_same()

        return progress

    def _next_nonsame_page(self, i):
        if i >= len(self.page_list):
            return None, i
        page = self.page_list[i]
        while page.same:
            i += 1
            if i >= len(self.page_list):
                return None, i
            page = self.page_list[i]
        return page, i + 1

    def _sector_erase_program_double_buffer(self, progress_cb=_stub_progress):
        """@brief Double-buffered program by performing sector erases."""
        actual_sector_erase_count = 0
        actual_sector_erase_weight = 0
        progress = 0

        progress_cb(0.0)

        program_timeout = self.flash.target.session.options.get('flash.timeout.program')

        # Fill in same flag for all pages. This is done up front so we're not trying
        # to read from flash while simultaneously programming it.
        progress = self._scan_pages_for_same(progress_cb)

        if self.region.is_erasable:
            # Erase all sectors up front.
            self.flash.init(self.flash.Operation.ERASE)
            for sector in self.sector_list:
                if sector.are_any_pages_not_same():
                    # Erase the sector
                    for addr in sector.addrs:
                        self.flash.erase_sector(addr)

                    # Update progress
                    progress += sector.erase_weight
                    if self.sector_erase_weight > 0:
                        progress_cb(float(progress) / float(self.sector_erase_weight))
            self.flash.uninit()

        # Set up page and buffer info.
        current_buf = 0
        next_buf = 1
        page, i = self._next_nonsame_page(0)

        # Make sure there are actually pages to program differently from current flash contents.
        if page is not None:
            self.flash.init(self.flash.Operation.PROGRAM)

            # Load first page buffer
            self.flash.load_page_buffer(current_buf, page.addr, page.data)

            while page is not None:
                assert page.same is not None

                # Kick off this page program.
                current_addr = page.addr
                current_weight = page.get_program_weight()

                self.flash.start_program_page_with_buffer(current_buf, current_addr)

                actual_sector_erase_count += 1
                actual_sector_erase_weight += page.get_program_weight()

                # Get next page and load it.
                page, i = self._next_nonsame_page(i)
                if page is not None:
                    self.flash.load_page_buffer(next_buf, page.addr, page.data)

                # Wait for the program to complete.
                result = self.flash.wait_for_completion(timeout=program_timeout)
                if result == self.flash.TIMEOUT_ERROR:
                    raise FlashProgramFailure('flash program page timeout', address=current_addr, result_code=result)
                elif result != 0:
                    raise FlashProgramFailure('flash program page failure', address=current_addr, result_code=result)

                # Swap buffers.
                current_buf, next_buf = next_buf, current_buf

                # Update progress
                progress += current_weight
                if self.sector_erase_weight > 0:
                    progress_cb(float(progress) / float(self.sector_erase_weight))

            self.flash.uninit()

        progress_cb(1.0)

        LOG.debug("Estimated sector erase programmed page count: %i", self.sector_erase_count)
        LOG.debug("Actual sector erase programmed page count: %i", actual_sector_erase_count)

        return FlashBuilder.FLASH_SECTOR_ERASE



# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2022 NXP
# Copyright (c) 2022-2023 Chris Reed
# Copyright (c) 2023 Hardy Griech
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import io
import logging
import time
import datetime
import pathlib
from enum import (Enum, IntEnum)
from typing import (IO, TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union, cast, overload)
from typing_extensions import Literal

from ..coresight.cortex_m import CortexM
from ..core import (exceptions, session)

if TYPE_CHECKING:
    from .context import DebugContext

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

## bkpt #0xab instruction
BKPT_INSTR = 0xbeab

class SemihostingRequests(IntEnum):
    """@brief Arm semihosting request numbers."""
    SYS_OPEN            = 0x01
    SYS_CLOSE           = 0x02
    SYS_WRITEC          = 0x03
    SYS_WRITE0          = 0x04
    SYS_WRITE           = 0x05
    SYS_READ            = 0x06
    SYS_READC           = 0x07
    SYS_ISERROR         = 0x08
    SYS_ISTTY           = 0x09
    SYS_SEEK            = 0x0a
    SYS_FLEN            = 0x0c
    SYS_TMPNAM          = 0x0d
    SYS_REMOVE          = 0x0e
    SYS_RENAME          = 0x0f
    SYS_CLOCK           = 0x10
    SYS_TIME            = 0x11
    SYS_SYSTEM          = 0x12
    SYS_ERRNO           = 0x13
    SYS_GET_CMDLINE     = 0x15
    SYS_HEAPINFO        = 0x16
    angel_SWIreason_EnterSVC = 0x17 # pylint: disable=invalid-name
    SYS_EXIT            = 0x18 # Also called angel_SWIreason_ReportException
    SYS_EXIT_EXTENDED   = 0x20
    SYS_ELAPSED         = 0x30
    SYS_TICKFREQ        = 0x31

# Pseudo-file descriptor numbers.
# Note: According to Arm semihosting spec, the fds must be non-zero.  But to achive POSIX compatibility
#       it has been chosen to use 0 for STDIN_FD.  OpenOCD behaves the same.
STDIN_FD = 0
STDOUT_FD = 1
STDERR_FD = 2

## Maximum length of a null-terminated string we'll attempt to read from target memory.
#
# The length is limited in case the string isn't terminated, this is relevant only for TARGET_SYS_WRITE0
#
# @see SemihostAgent::get_data()
MAX_STRING_LENGTH = 2048

## Enumsused for the file ID to indicate a special file was opened.
class SpecialFile(Enum):
    # ":semihosting-features"
    SEMIHOSTING_FEATURES_FILE = object()

class SemihostIOHandler:
    """@brief Interface for semihosting file I/O handlers.

    This class is also used as the default I/O handler if none is provided to SemihostAgent.
    In this case, all file I/O requests are rejected.
    """

    agent: Optional["SemihostAgent"]
    _errno: int

    def __init__(self) -> None:
        self.agent = None
        self._errno = 0

    def cleanup(self) -> None:
        pass

    @property
    def errno(self) -> int:
        return self._errno

    def _std_open(self, fnptr: int, fnlen: int, mode: str) -> Tuple[Optional[Union[int, SpecialFile]], str]:
        """@brief Helper for standard I/O open requests.

        In the Arm semihosting spec, standard I/O files are opened using a filename of ":tt"
        with the open mode specifying which standard I/O file to open. This method takes care
        of these special open requests, and is intended to be used by concrete I/O handler
        subclasses.

        Another special file is the ":semihosting-features" file used for semihosting feature bit
        reporting. This method recognised this file name, checks the requested file mode against
        allowed modes, and returns `SpecialFile.SEMIHOSTING_FEATURES_FILE` as the file ID.

        @return A 2-tuple of the file descriptor and filename. The filename is returned so it
          only has to be read from target memory once if the request is not for standard I/O.
          The returned file descriptor may be one of 0, 1, or 2 for the standard I/O files,
          -1 if an invalid combination was requested, or None if the request was not for
          a standard I/O file (i.e., the filename was not ":tt"). If None is returned for the
          file descriptor, the caller must handle the open request.
          SpecialFile.SEMIHOSTING_FEATURES_FILE can also be returned as the file ID in case
          the special ":semihosting-features" file is opened.
        @exception IOError Raised if an invalid file mode is used for a special file.
        """
        assert self.agent
        filename = self.agent.get_data(fnptr, fnlen).decode()
        LOG.debug("Semihost: open '%s' mode %s", filename, mode)

        # Handle standard I/O.
        if filename == ':tt':
            if mode == 'r':
                fd = STDIN_FD
            elif mode == 'w':
                fd = STDOUT_FD
            elif mode == 'a':
                fd = STDERR_FD
            else:
                LOG.warning("Unrecognized semihosting console open file combination: mode=%s", mode)
                return -1, filename
            return fd, filename
        # Semihosting features file, not currently supported.
        elif filename == ':semihosting-features':
            # All modes other than 'r' and 'rb' must fail.
            if mode not in ('r', 'rb'):
                raise IOError("attempt to open :semihosting-features with invalid mode "
                                "(only r and rb are allowed)")
            return SpecialFile.SEMIHOSTING_FEATURES_FILE, filename
        return None, filename

    def open(self, fnptr: int, fnlen: int, mode: str) -> int:
        raise NotImplementedError()

    def close(self, fd: int) -> int:
        raise NotImplementedError()

    def write(self, fd: int, ptr: int, length: int) -> int:
        raise NotImplementedError()

    def read(self, fd: int, ptr: int, length: int) -> int:
        raise NotImplementedError()

    def readc(self) -> int:
        raise NotImplementedError()

    def istty(self, fd: int) -> int:
        raise NotImplementedError()

    def seek(self, fd: int, pos: int) -> int:
        raise NotImplementedError()

    def flen(self, fd: int) -> int:
        raise NotImplementedError()

    def remove(self, ptr: int, length: int) -> int:
        raise NotImplementedError()

    def rename(self, oldptr: int, oldlength: int, newptr: int, newlength: int) -> int:
        raise NotImplementedError()

class InternalSemihostIOHandler(SemihostIOHandler):
    """@brief Implements semihosting requests directly in the Python process.

    This class maintains its own list of pseudo-file descriptors for files opened by the
    debug target. By default, this class uses the system stdin, stdout, and stderr file objects
    for file desscriptors 1, 2, and 3.
    """

    def __init__(self):
        super().__init__()
        self.next_fd = STDERR_FD + 1

        # Go ahead and connect standard I/O.
        self.open_files: Dict[int, Union[IO[str], IO[bytes]]] = {
                STDIN_FD : sys.stdin,
                STDOUT_FD : sys.stdout,
                STDERR_FD : sys.stderr
            }

    def _is_valid_fd(self, fd):
        return fd in self.open_files and self.open_files[fd] is not None

    def cleanup(self):
        for f in (self.open_files[k] for k in self.open_files if k > STDERR_FD):
            f.close()

    def open(self, fnptr, fnlen, mode):
        special_fd, filename = self._std_open(fnptr, fnlen, mode)
        # if special_fd is not None:
        #     return special_fd
        if (special_fd is not None) and (special_fd is not SpecialFile.SEMIHOSTING_FEATURES_FILE):
            return special_fd

        try:
            # Handle semihosting features.
            if special_fd is SpecialFile.SEMIHOSTING_FEATURES_FILE:
                # Features bits:
                # - Byte 0, bit 0 = 0: SH_EXT_EXIT_EXTENDED, whether SYS_EXIT_EXTENDED is supported
                # - Byte 0, bit 1 = 1: SH_EXT_STDOUT_STDERR, whether both stdout and stderr are supported
                f = io.BytesIO(b"SHFB\x02")
            else:
                # Expand user directory.
                filepath = pathlib.Path(filename).expanduser()

                # ensure directories are exists if mode is write/appened
                if ('w' in mode) or ('a' in mode):
                    filepath.parent.mkdir(parents=True, exist_ok=True)

                f = io.open(filepath, mode)

            fd = self.next_fd
            self.next_fd += 1

            self.open_files[fd] = f

            return fd
        except OSError as e:
            self._errno = e.errno
            LOG.error("Semihost: failed to open file '%s'", filename,
                    exc_info=(self.agent.context.session.log_tracebacks if self.agent else True))
            return -1

    def close(self, fd):
        if fd > STDERR_FD:
            if not self._is_valid_fd(fd):
                return -1
            f = self.open_files.pop(fd)
            try:
                f.close()
            except OSError:
                # Ignore errors closing files.
                pass
        return 0

    def write(self, fd, ptr, length):
        if not self._is_valid_fd(fd):
            # Return byte count not written.
            return length

        assert self.agent
        data = self.agent.get_data(ptr, length)
        f = self.open_files[fd]
        try:
            if 'b' in f.mode:
                cast(IO[bytes], f).write(data)
            else:
                cast(IO[str], f).write(data.decode(errors='ignore'))
            f.flush()
            return 0
        except OSError as e:
            self._errno = e.errno
            LOG.debug("Semihost: exception: %s", e)
            return -1

    def read(self, fd, ptr, length):
        assert self.agent

        if not self._is_valid_fd(fd):
            # Return byte count not read.
            return length

        try:
            f = self.open_files[fd]
            data = f.read(length)
        except OSError as e:
            self._errno = e.errno
            LOG.debug("Semihost: exception: %s", e)
            return -1

        if isinstance(data, str):
            data = data.encode()

        ba = bytearray(data)
        self.agent.context.write_memory_block8(ptr, ba)
        return length - len(ba)

    def readc(self):
        try:
            f = self.open_files[STDIN_FD]
            if f is not None:
                data = f.read(1)
                c = ord(data[0:1])
                return c
            else:
                return 0
        except OSError as e:
            self._errno = e.errno
            return 0

    def istty(self, fd):
        if not self._is_valid_fd(fd):
            return -1
        # Just assume that stdio is a terminal and other files aren't.
        return int(not fd > STDERR_FD)

    def seek(self, fd, pos):
        if not self._is_valid_fd(fd):
            return -1
        try:
            self.open_files[fd].seek(pos)
            return 0
        except OSError as e:
            self._errno = e.errno
            return -1

    def flen(self, fd):
        if not self._is_valid_fd(fd):
            return -1
        f = self.open_files[fd]
        try:
            info = os.fstat(f.fileno())
            return info.st_size
        except io.UnsupportedOperation:
            # Try seeking to end to get size.
            saved_pos = f.tell()
            f.seek(0, os.SEEK_END)
            size = f.tell()
            f.seek(saved_pos, os.SEEK_SET)
            return size
        except OSError as e:
            self._errno = e.errno
            return -1

class ConsoleIOHandler(SemihostIOHandler):
    """@brief Simple IO handler for console."""

    def __init__(self, stdin_file, stdout_file=None):
        super().__init__()
        self._stdin_file = stdin_file
        self._stdout_file = stdout_file or stdin_file

    def write(self, fd, ptr, length):
        assert self.agent
        data = self.agent.get_data(ptr, length)
        self._stdout_file.write(data)
        return 0

    def read(self, fd, ptr, length):
        assert self.agent
        data = self._stdin_file.read(length)

        # Stuff data into provided buffer.
        if data:
            self.agent.context.write_memory_block8(ptr, data)

        result = length - len(data)
        if not data:
            self._errno = 5
            return -1
        return result

    def readc(self):
        data = self._stdin_file.read(1)

        if data:
            return ord(data)
        else:
            return -1

class SemihostAgent:
    """@brief Handler for Arm semihosting requests.

    Semihosting requests are made by the target by executing a 'bkpt #0xab' instruction. The
    requested operation is specified by R0 and any arguments by R1. Many requests use a block
    of word-sized arguments pointed to by R1. The return value is passed back to the target
    in R0.

    This class does not handle any file-related requests by itself. It uses I/O handler objects
    passed in to the constructor. The requests handled directly by this class are SYS_CLOCK and
    SYS_TIME.

    There are two types of I/O handlers used by this class. The main I/O handler, set
    with the constructor's @i io_handler parameter, is used for most file operations.
    You may optionally pass another I/O handler for the @i console constructor parameter. The
    console handler is used solely for standard I/O and debug console I/O requests. If no console
    handler is provided, the main handler is used instead. SYS_OPEN requests are not
    passed to the console handler in any event, they are always passed to the main handler.

    If no main I/O handler is provided, the class will use SemihostIOHandler, which causes all
    file I/O requests to be rejected as an error.

    The SemihostAgent assumes standard I/O file descriptor numbers are #STDIN_FD, #STDOUT_FD,
    and #STDERR_FD. When it receives a read or write request for one of these descriptors, it
    passes the request to the console handler. This means the main handler must return these
    numbers for standard I/O open requests (those with a file name of ":tt").

    Not all semihosting requests are supported. Those that are not implemented are:
    - SYS_TMPNAM
    - SYS_SYSTEM
    - SYS_GET_CMDLINE
    - SYS_HEAPINFO
    - SYS_EXIT
    - SYS_ELAPSED
    - SYS_TICKFREQ
    """

    ## Index into this array is the file open mode argument to SYS_OPEN.
    OPEN_MODES = ['r', 'rb', 'r+', 'r+b', 'w', 'wb', 'w+', 'w+b', 'a', 'ab', 'a+', 'a+b']

    EPOCH = datetime.datetime(1970, 1, 1)

    def __init__(
            self,
            context: "DebugContext",
            io_handler: Optional[SemihostIOHandler] = None,
            console: Optional[SemihostIOHandler] = None
        ) -> None:
        self.context = context
        self.start_time = time.time()
        self.io_handler = io_handler or SemihostIOHandler()
        self.io_handler.agent = self
        self.console = console or self.io_handler
        self.console.agent = self

    def check_and_handle_semihost_request(self) -> bool:
        """@brief Handle a semihosting request.

        This method should be called after the target has halted, to check if the halt was
        due to a semihosting request. It first checks to see if the target halted because
        of a breakpoint. If so, it reads the instruction at PC to make sure it is a 'bkpt #0xAB'
        instruction. If so, the target is making a semihosting request. If not, nothing more is done.

        After the request is handled, the PC is advanced to the next instruction after the 'bkpt'.
        A boolean is return indicating whether a semihosting request was handled. If True, the
        caller should resume the target immediately.

        @retval True A semihosting request was handled.
        @retval False The target halted for a reason other than semihosting, i.e. a user-installed
          debugging breakpoint.
        """
        # Nothing to do if this is not a bkpt.
        if (self.context.read32(CortexM.DFSR) & CortexM.DFSR_BKPT) == 0:
            return False

        pc = self.context.read_core_register('pc')
        assert isinstance(pc, int)

        # Are we stopped due to one of our own breakpoints?
        # TODO check against watchpoints too!?
        bp = self.context.core.find_breakpoint(pc)
        if bp:
            return False

        # Get the instruction at the breakpoint.
        instr = self.context.read16(pc)

        # Check for semihost bkpt.
        if instr != BKPT_INSTR:
            return False

        # Advance PC beyond the bkpt instruction.
        self.context.write_core_register('pc', pc + 2)

        # Get args
        op = self.context.read_core_register('r0')
        args = self.context.read_core_register('r1')
        assert isinstance(op, int)
        assert isinstance(args, int)

        # Handle request
        handler = self._REQUEST_MAP.get(op, None)
        if handler:
            try:
                result = handler(self, args)
            except NotImplementedError:
                LOG.warning("Semihost: unimplemented request pc=%x r0=%x r1=%x", pc, op, args)
                result = -1
            except (exceptions.Error, OSError) as e:
                LOG.error("Error while handling semihost request: %s", e,
                    exc_info=self.context.session.log_tracebacks)
                result = -1
        else:
            result = -1

        # Set return value.
        self.context.write_core_register('r0', result)

        return True

    def cleanup(self) -> None:
        """@brief Clean up any resources allocated by semihost requests.

        @note May be called more than once.
        """
        self.io_handler.cleanup()
        if self.console is not self.io_handler:
            self.console.cleanup()

    @overload
    def _get_args(self, args_address: int, count: Literal[1]) -> int:
        ...

    @overload
    def _get_args(self, args_address: int, count: int) -> List[int]:
        ...

    def _get_args(self, args_address: int, count):
        args = self.context.read_memory_block32(args_address, count)
        if count == 1:
            return args[0]
        else:
            return args

    def get_data(self, ptr: int, length: Optional[int] = None) -> bytes:
        if length is not None:
            data = self.context.read_memory_block8(ptr, length)
            return bytes(data)

        target_data = b''
        data = b''
        # TODO - use memory map to make sure we don't try to read off the end of memory
        # Limit string size in case it isn't terminated.
        while len(target_data) < MAX_STRING_LENGTH:
            try:
                # Read 32 bytes at a time for efficiency.
                data = self.context.read_memory_block8(ptr, 32)
                terminator = data.index(0)

                # Found a null terminator, append data up to but not including the null
                # and then exit the loop.
                target_data += bytes(data[:terminator])
                break
            except exceptions.TransferError:
                # Failed to read some or all of the string.
                break
            except ValueError:
                # No null terminator was found. Append all of data.
                target_data += bytes(data)
                ptr += 32
        return target_data

    def handle_sys_open(self, args: int) -> int:
        fnptr, mode, fnlen = self._get_args(args, 3)
        if mode >= len(self.OPEN_MODES):
            return -1
        mode = self.OPEN_MODES[mode]

        TRACE.debug("Semihost: open %x/%x, mode %s", fnptr, fnlen, mode)
        return self.io_handler.open(fnptr, fnlen, mode)

    def handle_sys_close(self, args: int) -> int:
        fd = self._get_args(args, 1)
        TRACE.debug("Semihost: close fd=%d", fd)
        return self.io_handler.close(fd)

    def handle_sys_writec(self, args: int) -> int:
        TRACE.debug("Semihost: writec %x", args)
        return self.console.write(STDOUT_FD, args, 1)

    def handle_sys_write0(self, args: int) -> int:
        msg = self.get_data(args)
        TRACE.debug("Semihost: write0 msg='%s'", msg)
        return self.console.write(STDOUT_FD, args, len(msg))

    def handle_sys_write(self, args: int) -> int:
        fd, data_ptr, length = self._get_args(args, 3)
        TRACE.debug("Semihost: write fd=%d ptr=%x len=%d", fd, data_ptr, length)
        if fd in (STDOUT_FD, STDERR_FD):
            return self.console.write(fd, data_ptr, length)
        else:
            return self.io_handler.write(fd, data_ptr, length)

    def handle_sys_read(self, args: int) -> int:
        fd, ptr, length = self._get_args(args, 3)
        TRACE.debug("Semihost: read fd=%d ptr=%x len=%d", fd, ptr, length)
        if fd == STDIN_FD:
            return self.console.read(fd, ptr, length)
        else:
            return self.io_handler.read(fd, ptr, length)

    def handle_sys_readc(self, args: int) -> int:
        TRACE.debug("Semihost: readc")
        return self.console.readc()

    def handle_sys_iserror(self, args: int) -> int:
        raise NotImplementedError()

    def handle_sys_istty(self, args: int) -> int:
        fd = self._get_args(args, 1)
        TRACE.debug("Semihost: istty fd=%d", fd)
        return self.io_handler.istty(fd)

    def handle_sys_seek(self, args: int) -> int:
        fd, pos = self._get_args(args, 2)
        TRACE.debug("Semihost: seek fd=%d pos=%d", fd, pos)
        return self.io_handler.seek(fd, pos)

    def handle_sys_flen(self, args: int) -> int:
        fd = self._get_args(args, 1)
        TRACE.debug("Semihost: flen fd=%d", fd)
        return self.io_handler.flen(fd)

    def handle_sys_tmpnam(self, args: int) -> int:
        raise NotImplementedError()

    def handle_sys_remove(self, args: int) -> int:
        ptr, length = self._get_args(args, 2)
        return self.io_handler.remove(ptr, length)

    def handle_sys_rename(self, args: int) -> int:
        oldptr, oldlength, newptr, newlength = self._get_args(args, 4)
        return self.io_handler.rename(oldptr, oldlength, newptr, newlength)

    def handle_sys_clock(self, args: int) -> int:
        now = time.time()
        delta = now - self.start_time
        return int(delta * 100)

    def handle_sys_time(self, args: int) -> int:
        now = datetime.datetime.now()
        delta = now - self.EPOCH
        seconds = (delta.days * 86400) + delta.seconds
        return seconds

    def handle_sys_system(self, args: int) -> int:
        raise NotImplementedError()

    def handle_sys_errno(self, args: int) -> int:
        return self.io_handler.errno

    def handle_sys_get_cmdline(self, args: int) -> int:
        cmdline = cast(str, self.context.session.options.get('semihost.commandline'))
        if not cmdline:
            return -1

        ptr, length = self._get_args(args, 2)
        cmdline_write_length = min(length - 1, len(cmdline)) # Ensure room for null byte.
        cmdline_bytes = cmdline.encode()[:cmdline_write_length] + b'\x00'
        self.context.write_memory_block8(ptr, cmdline_bytes)
        self.context.write32(args + 4, cmdline_write_length - 1) # TODO resume assumption about pointer size!
        return 0

    def handle_sys_heapinfo(self, args: int) -> int:
        """@brief Stub implementation of SYS_HEAPINFO.

        The args (r1) value is the address of a pointer to a four-word data block, to be filled in
        by the host.

        ```c
        struct block {
            int heap_base;
            int heap_limit;
            int stack_base;
            int stack_limit;
        };
        ```

        This implementation simply fills in the value 0 for each field. Zero is legal, and tells the
        caller that the host was unable to determine the value.
        """
        info_block = self._get_args(args, 1)
        self.context.write_memory_block32(info_block, [0, 0, 0, 0])
        return 0

    def handle_sys_exit(self, args: int) -> int:
        # TODO handle SYS_EXIT for a 'pyocd run' subcommand
        raise NotImplementedError()

    def handle_sys_exit_extended(self, args: int) -> int:
        raise NotImplementedError()

    def handle_sys_elapsed(self, args: int) -> int:
        raise NotImplementedError()

    def handle_sys_tickfreq(self, args: int) -> int:
        raise NotImplementedError()

    _REQUEST_MAP: Dict[int, Callable[["SemihostAgent", int], int]] = {
            SemihostingRequests.SYS_OPEN:            handle_sys_open,
            SemihostingRequests.SYS_CLOSE:           handle_sys_close,
            SemihostingRequests.SYS_WRITEC:          handle_sys_writec,
            SemihostingRequests.SYS_WRITE0:          handle_sys_write0,
            SemihostingRequests.SYS_WRITE:           handle_sys_write,
            SemihostingRequests.SYS_READ:            handle_sys_read,
            SemihostingRequests.SYS_READC:           handle_sys_readc,
            SemihostingRequests.SYS_ISERROR:         handle_sys_iserror,
            SemihostingRequests.SYS_ISTTY:           handle_sys_istty,
            SemihostingRequests.SYS_SEEK:            handle_sys_seek,
            SemihostingRequests.SYS_FLEN:            handle_sys_flen,
            SemihostingRequests.SYS_TMPNAM:          handle_sys_tmpnam,
            SemihostingRequests.SYS_REMOVE:          handle_sys_remove,
            SemihostingRequests.SYS_RENAME:          handle_sys_rename,
            SemihostingRequests.SYS_CLOCK:           handle_sys_clock,
            SemihostingRequests.SYS_TIME:            handle_sys_time,
            SemihostingRequests.SYS_SYSTEM:          handle_sys_system,
            SemihostingRequests.SYS_ERRNO:           handle_sys_errno,
            SemihostingRequests.SYS_GET_CMDLINE:     handle_sys_get_cmdline,
            SemihostingRequests.SYS_HEAPINFO:        handle_sys_heapinfo,
            SemihostingRequests.SYS_EXIT:            handle_sys_exit,
            SemihostingRequests.SYS_EXIT_EXTENDED:   handle_sys_exit_extended,
            SemihostingRequests.SYS_ELAPSED:         handle_sys_elapsed,
            SemihostingRequests.SYS_TICKFREQ:        handle_sys_tickfreq
        }




# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import (Optional, List)
import logging
import sys
import os
from time import sleep

from .base import SubcommandBase
from ..core.helpers import ConnectHelper
from ..core.session import Session
from ..utility.cmdline import (
    convert_session_options,
    split_command_line,
    )
from ..probe.shared_probe_proxy import SharedDebugProbeProxy
from ..gdbserver import GDBServer
from ..probe.tcp_probe_server import DebugProbeServer
from ..coresight.generic_mem_ap import GenericMemAPTarget
from ..utility.notification import Notification

LOG = logging.getLogger(__name__)

class GdbserverSubcommand(SubcommandBase):
    """@brief `pyocd gdbserver` subcommand."""

    NAMES = ['gdbserver', 'gdb']
    HELP = "Run the gdb remote server(s)."

    ## @brief Valid erase mode options.
    ERASE_OPTIONS = [
        'auto',
        'chip',
        'sector',
        ]

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""
        gdbserver_parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        gdbserver_options = gdbserver_parser.add_argument_group("gdbserver options")
        gdbserver_options.add_argument("-p", "--port", metavar="PORT", dest="port_number", type=int,
            default=3333,
            help="Set starting port number for the GDB server (default 3333). Additional cores "
                "will have a port number of this parameter plus the core number.")
        gdbserver_options.add_argument("-T", "--telnet-port", metavar="PORT", dest="telnet_port",
            type=int, default=4444,
            help="Specify starting telnet port for semihosting (default 4444).")
        gdbserver_options.add_argument("-R", "--probe-server-port",
            dest="probe_server_port", metavar="PORT", type=int, default=5555,
            help="Specify the port for the probe server (default 5555).")
        gdbserver_options.add_argument("-r", "--probe-server", action="store_true", dest="enable_probe_server",
            help="Enable the probe server in addition to the GDB server.")
        gdbserver_options.add_argument("--allow-remote", dest="serve_local_only", default=True, action="store_false",
            help="Allow remote TCP/IP connections (default is no).")
        gdbserver_options.add_argument("--persist", action="store_true",
            help="Keep GDB server running even after remote has detached.")
        gdbserver_options.add_argument("--core", metavar="CORE_LIST",
            help="Comma-separated list of core numbers for which gdbservers will be created. Default is all cores.")
        gdbserver_options.add_argument("--elf", metavar="PATH",
            help="Optionally specify ELF file being debugged.")
        gdbserver_options.add_argument("-e", "--erase", choices=cls.ERASE_OPTIONS, default='sector',
            help="Choose flash erase method. Default is sector.")
        gdbserver_options.add_argument("--trust-crc", action="store_true",
            help="Use only the CRC of each page to determine if it already has the same data.")
        gdbserver_options.add_argument("-C", "--vector-catch", default='h',
            help="Enable vector catch sources, one letter per enabled source in any order, or 'all' "
                "or 'none'. (b=bus fault, c=check err, e=secure fault, h=hard fault, i=irq err, m=mem fault, "
                "p=nocp, r=reset, s=state err, a=all, n=none). Default is hard fault.")
        gdbserver_options.add_argument("-S", "--semihosting", dest="enable_semihosting", action="store_true",
            help="Enable semihosting.")
        gdbserver_options.add_argument("--step-into-interrupts", dest="step_into_interrupt", default=False, action="store_true",
            help="Allow single stepping to step into interrupts.")
        gdbserver_options.add_argument("-c", "--command", dest="commands", metavar="CMD", action='append', nargs='+',
            help="Run command (OpenOCD compatibility).")

        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, gdbserver_parser]

    def __init__(self, args: argparse.Namespace):
        """@brief Constructor."""
        super().__init__(args)
        self._echo_msg = None

    def _process_commands(self, commands: Optional[List[str]]):
        """@brief Handle OpenOCD commands for compatibility."""
        if commands is None:
            return
        for cmd_list in commands:
            try:
                cmd_list = split_command_line(cmd_list)
                cmd = cmd_list[0]
                if cmd == 'gdb_port':
                    if len(cmd_list) < 2:
                        LOG.error("Missing port argument")
                    else:
                        self._args.port_number = int(cmd_list[1], base=0)
                elif cmd == 'telnet_port':
                    if len(cmd_list) < 2:
                        LOG.error("Missing port argument")
                    else:
                        self._args.telnet_port = int(cmd_list[1], base=0)
                elif cmd == 'echo':
                    self._echo_msg = ' '.join(cmd_list[1:])
                else:
                    LOG.error("Unsupported command: %s" % ' '.join(cmd_list))
            except IndexError:
                pass

    def _gdbserver_listening_cb(self, note: Notification):
        """@brief Callback invoked when the gdbserver starts listening on its port."""
        if self._echo_msg is not None:
            print(self._echo_msg, file=sys.stderr)
            sys.stderr.flush()

    def invoke(self) -> int:
        """@brief Handle 'gdbserver' subcommand."""
        self._process_commands(self._args.commands)

        probe_server = None
        gdbs = []
        try:
            # Build dict of session options.
            sessionOptions = convert_session_options(self._args.options)
            sessionOptions.update({
                'gdbserver_port' : self._args.port_number,
                'telnet_port' : self._args.telnet_port,
                'persist' : self._args.persist,
                'step_into_interrupt' : self._args.step_into_interrupt,
                'chip_erase': self._args.erase,
                'fast_program' : self._args.trust_crc,
                'enable_semihosting' : self._args.enable_semihosting,
                'serve_local_only' : self._args.serve_local_only,
                'vector_catch' : self._args.vector_catch,
                })

            # Split list of cores to serve.
            if self._args.core is not None:
                try:
                    core_list = {int(x) for x in self._args.core.split(',')}
                except ValueError as err:
                    LOG.error("Invalid value passed to --core")
                    return 1
            else:
                core_list = None

            # Get the probe.
            probe = ConnectHelper.choose_probe(
                        blocking=(not self._args.no_wait),
                        return_first=False,
                        unique_id=self._args.unique_id,
                        )
            if probe is None:
                LOG.error("No probe selected.")
                return 1

            # Create a proxy so the probe can be shared between the session and probe server.
            probe_proxy = SharedDebugProbeProxy(probe)

            # Create the session.
            session = Session(probe_proxy,
                project_dir=self._args.project_dir,
                user_script=self._args.script,
                config_file=self._args.config,
                no_config=self._args.no_config,
                pack=self._args.pack,
                unique_id=self._args.unique_id,
                target_override=self._args.target_override,
                frequency=self._args.frequency,
                connect_mode=self._args.connect_mode,
                options=sessionOptions,
                option_defaults=self._modified_option_defaults(),
            )
            if session is None:
                LOG.error("No probe selected.")
                return 1
            with session:
                # Validate the core selection.
                all_cores = set(session.target.cores.keys())
                if core_list is None:
                    core_list = all_cores
                bad_cores = core_list.difference(all_cores)
                if len(bad_cores):
                    LOG.error("Invalid core number%s: %s",
                        "s" if len(bad_cores) > 1 else "",
                        ", ".join(str(x) for x in bad_cores))
                    return 1

                # Set ELF if provided.
                if self._args.elf:
                    session.board.target.elf = os.path.expanduser(self._args.elf)

                # Run the probe server is requested.
                if self._args.enable_probe_server:
                    probe_server = DebugProbeServer(session, session.probe,
                            self._args.probe_server_port, self._args.serve_local_only)
                    session.probeserver = probe_server
                    probe_server.start()

                # Start up the gdbservers.
                for core_number, core in session.board.target.cores.items():
                    # Don't create a server for CPU-less memory Access Port.
                    if isinstance(session.board.target.cores[core_number], GenericMemAPTarget):
                        continue
                    # Don't create a server if this core is not listed by the user.
                    if core_number not in core_list:
                        continue
                    gdb = GDBServer(session, core=core_number)
                    # Only subscribe to the server for the first core, so echo messages aren't printed
                    # multiple times.
                    if not gdbs:
                        session.subscribe(self._gdbserver_listening_cb, GDBServer.GDBSERVER_START_LISTENING_EVENT, gdb)
                    session.gdbservers[core_number] = gdb
                    gdbs.append(gdb)
                    gdb.start()

                while any(g.is_alive() for g in gdbs):
                    sleep(0.1)
                if probe_server:
                    probe_server.stop()
        except (KeyboardInterrupt, Exception):
            for server in gdbs:
                server.stop()
            if probe_server:
                probe_server.stop()
            raise

        return 0




# pyOCD debugger
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from time import sleep
from typing import (cast, Dict, TYPE_CHECKING)

from ...core import exceptions
from ...coresight.coresight_target import CoreSightTarget
from ...coresight.ap import (APAddressBase, APv1Address, AccessPort, MEM_AP)
from ...probe.debug_probe import DebugProbe
from .delegates import DebugSequenceFunctionsDelegate
from .sequences import DebugSequenceRuntimeError

if TYPE_CHECKING:
    from ...coresight.dap import DebugPort

LOG = logging.getLogger(__name__)

# Disable warnings for the non-standard methods names we use to match the sequences function
# names, since introspection is used to look up functions.
# pylint: disable=invalid_name

class DebugSequenceCommonFunctions(DebugSequenceFunctionsDelegate):
    """@brief Implements functions provided by the debug sequence environment."""

    APSEL_SHIFT = 24

    DP_ABORT = 0x00

    def __init__(self) -> None:
        self._ap_cache: Dict[APAddressBase, MEM_AP] = {}

    @property
    def target(self) -> CoreSightTarget:
        return cast(CoreSightTarget, self.context.session.target)

    def _get_ap_addr(self) -> APAddressBase:
        """@brief Return the AP address selected by __ap or __apid variables.

        If the DFP uses apids then the `__apid` variable takes precedence. Otherwise the `__ap` variable
        is used.
        """
        pack_device = self.context.delegate.cmsis_pack_device

        if pack_device.uses_apid:
            apid = self.context.get_variable('__apid')
            try:
                ap_addr = pack_device.apid_map[apid]
            except KeyError:
                raise DebugSequenceRuntimeError(f"__apid has invalid value ({apid})")
        else:
            # The __ap variable is only used to reference v1 APs.
            # TODO handle __dp being non-zero, when we support multiple DPs.
            ap_addr = APv1Address(self.context.get_variable('__ap'))

            dp_num = self.context.get_variable('__dp')
            if dp_num != 0:
                raise DebugSequenceRuntimeError(f"currently only __dp==0 is supported ({dp_num} specified)")

        return ap_addr

    def _get_mem_ap(self) -> MEM_AP:
        """@brief Return the current MEM_AP object.

        Normally, the AP object created during discovery is returned from the target's dict of APs. However,
        sequences can be run prior to discovery, and are allowed to perform memory acceses. Therefore we must
        handle the case of there not being a readily available AP object by creating a temporary one here.

        A cache dict is used to prevent repeatedly recreating the same AP when multiple memory transfers
        appear in a sequence. The cache is _only_ used for temporary AP objects; the target's AP dict always
        takes priority and is checked first.
        """
        ap_addr = self._get_ap_addr()

        # Try to get an existing AP from the target.
        try:
            ap = self.target.aps[ap_addr]
            if not isinstance(ap, MEM_AP):
                raise DebugSequenceRuntimeError(f"AP at address {ap_addr} is not a MEM-AP")
            return ap
        except KeyError:
            pass

        # The AP doesn't exist or we haven't performed discovery yet, but we still need to support memory
        # transfers for debug sequences. So attempt to create a temporary one.
        # TODO can this only be done prior to discovery?

        # Check if we have already created and cached this AP.
        try:
            return self._ap_cache[ap_addr]
        except KeyError:
            pass

        # Haven't encountered this AP yet. Create and cache it.
        # This call will raise exceptions.TargetError if there is no AP with the requested address.
        ap = AccessPort.create(self.target.dp, ap_addr)

        # Make sure this is a MEM-AP.
        if not isinstance(ap, MEM_AP):
            raise DebugSequenceRuntimeError(f"AP at address {ap_addr} is not a MEM-AP")

        # Save in the cache.
        self._ap_cache[ap_addr] = ap

        return ap

    def _get_dp(self, ignore_apid: bool = False) -> DebugPort:
        """@brief Get the DebugPort object specified by the __dp or __apid variable."""
        pack_device = self.context.delegate.cmsis_pack_device
        if not pack_device.uses_apid or ignore_apid:
            dp_num = self.context.get_variable('__dp')
        else:
            apid = self.context.get_variable('__apid')
            try:
                ap_addr = pack_device.apid_map[apid]
            except KeyError:
                raise DebugSequenceRuntimeError(f"__apid has invalid value ({apid})")
            else:
                dp_num = ap_addr.dp_index

        if dp_num != 0:
            raise DebugSequenceRuntimeError(f"currently only __dp==0 is supported ({dp_num} specified)")

        # In any case, for now we always return the only DebugPort object we have.
        return self.target.dp

    def _get_ignore_errors(self) -> bool:
        """@brief Whether the debug sequence has set __errorcontrol to ignore faults."""
        errcontrol = self.context.get_variable('__errorcontrol')
        return (errcontrol & 1) == 1

    def sequence(self, name: str) -> None:
        # This call will raise if the named sequence is invalid. However, we should have already
        # verified the sequence name is valid during semantic checking.
        #
        # The pname from the current context is passed in order to match a pname-specific
        # sequence; a matching sequence with no pname will also be found.
        seq = self.context.delegate.get_sequence_with_name(name, pname=self.context.pname)

        if self.context.pname:
            LOG.debug("Running debug sub-sequence '%s' (%s)", name, self.context.pname)
        else:
            LOG.debug("Running debug sub-sequence '%s'", name)

        # Run the sequence.
        subsequence_scope = seq.execute(self.context)

        # Copy the result to parent sequence.
        if subsequence_scope is not None:
            result_value = subsequence_scope.get('__Result')
            LOG.debug("Sub-sequence '%s' result = %d", name, result_value)
            self.context.current_scope.set('__Result', result_value)

    def read8(self, addr: int) -> int:
        try:
            return self._get_mem_ap().read8(addr)
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Read8(%#010x) ignored %r because __errorcontrol is set", addr, err)
                return 0
            else:
                raise

    def read16(self, addr: int) -> int:
        try:
            return self._get_mem_ap().read16(addr)
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Read16(%#010x) ignored %r because __errorcontrol is set", addr, err)
                return 0
            else:
                raise

    def read32(self, addr: int) -> int:
        try:
            return self._get_mem_ap().read32(addr)
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Read32(%#010x) ignored %r because __errorcontrol is set", addr, err)
                return 0
            else:
                raise

    def read64(self, addr: int) -> int:
        try:
            return self._get_mem_ap().read64(addr)
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Read64(%#010x) ignored %r because __errorcontrol is set", addr, err)
                return 0
            else:
                raise

    def readap(self, addr: int) -> int:
        try:
            ap_addr = self._get_ap_addr()
            reg_addr = ap_addr.address | addr
            return self._get_dp().read_ap(reg_addr)
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("ReadAP(%#010x) ignored %r because __errorcontrol is set", addr, err)
                return 0
            else:
                raise

    def readaccessap(self, addr: int) -> int:
        try:
            dp = self._get_dp(True)
            apacc = dp.apacc_memory_interface
            return apacc.read32(addr)
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("ReadAccessAP(%#010x) ignored %r because __errorcontrol is set", addr, err)
                return 0
            else:
                raise

    def readdp(self, addr: int) -> int:
        try:
            return self._get_dp(True).read_dp(addr)
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("ReadDP(%#010x) ignored %r because __errorcontrol is set", addr, err)
                return 0
            else:
                raise

    def write8(self, addr: int, val: int) -> None:
        try:
            self._get_mem_ap().write8(addr, val)
            self.target.flush()
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Write8(%#010x) ignored %r because __errorcontrol is set", addr, err)
            else:
                raise

    def write16(self, addr: int, val: int) -> None:
        try:
            self._get_mem_ap().write16(addr, val)
            self.target.flush()
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Write16(%#010x) ignored %r because __errorcontrol is set", addr, err)
            else:
                raise

    def write32(self, addr: int, val: int) -> None:
        try:
            self._get_mem_ap().write32(addr, val)
            self.target.flush()
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Write32(%#010x) ignored %r because __errorcontrol is set", addr, err)
            else:
                raise

    def write64(self, addr: int, val: int) -> None:
        try:
            self._get_mem_ap().write64(addr, val)
            self.target.flush()
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("Write64(%#010x) ignored %r because __errorcontrol is set", addr, err)
            else:
                raise

    def writeap(self, addr: int, val: int) -> None:
        try:
            ap_addr = self._get_ap_addr()
            reg_addr = ap_addr.address | addr
            self._get_dp().write_ap(reg_addr, val)
            self.target.flush()
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("WriteAP(%#010x) ignored %r because __errorcontrol is set", addr, err)
            else:
                raise

    def writeaccessap(self, addr: int, val: int) -> None:
        try:
            dp = self._get_dp(True)
            apacc = dp.apacc_memory_interface
            apacc.write32(addr, val)
            self.target.flush()
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("WriteAccessAP(%#010x) ignored %r because __errorcontrol is set", addr, err)
            else:
                raise

    def writedp(self, addr: int, val: int) -> None:
        try:
            self._get_dp(True).write_dp(addr, val)
            self.target.flush()
        except exceptions.TransferError as err:
            if self._get_ignore_errors():
                LOG.debug("WriteDP(%#010x) ignored %r because __errorcontrol is set", addr, err)
            else:
                raise

    def flashbufferwrite(self, addr: int, offset: int, length: int, mode: int) -> None:
        raise NotImplementedError()

    def dap_delay(self, delay: int) -> None:
        # Flush before sleeping, in case there are any outstanding transactions.
        self.target.flush()

        # TODO This is really expected to be sent to the target via the CMSIS-DAP command, but for most
        # cases this will work fine. However, it would fail for atomic sequences.
        sleep(delay / 1e6)

    def dap_writeabort(self, value: int) -> None:
        assert self.context.session.probe
        mode = self.context.session.probe.wire_protocol
        if mode == DebugProbe.Protocol.SWD:
            self._get_dp().write_reg(self.DP_ABORT, value)
        elif mode == DebugProbe.Protocol.JTAG:
            # TODO support jtag abort
            self._get_dp().write_reg(self.DP_ABORT, value)
        self.target.flush()

    def dap_swj_pins(self, pinout: int, pinselect: int, pinwait: int) -> int:
        """
        Pin bits:
        - Bit 0: SWCLK/TCK
        - Bit 1: SWDIO/TMS
        - Bit 2: TDI
        - Bit 3: TDO
        - Bit 5: nTRST
        - Bit 7: nRESET

        Must return 0xFFFFFFFF if the pins cannot be read.
        """
        from ...probe.cmsis_dap_probe import CMSISDAPProbe

        probe = self.context.session.probe
        assert probe
        if DebugProbe.Capability.PIN_ACCESS not in probe.capabilities:
            return 0xFFFFFFFF

        # Write pins if any were selected to be modified, wait if needed, then read
        # all available pins.
        if pinselect != 0:
            probe.write_pins(DebugProbe.PinGroup.PROTOCOL_PINS,
                    CMSISDAPProbe.from_cmsis_dap_pins(pinselect),
                    CMSISDAPProbe.from_cmsis_dap_pins(pinout))
        if pinwait > 0:
            sleep(pinwait / 1e9)
        result = CMSISDAPProbe.to_cmsis_dap_pins(
                probe.read_pins(DebugProbe.PinGroup.PROTOCOL_PINS,
                                DebugProbe.ProtocolPin.ALL_PINS))

        return result

    def dap_swj_clock(self, val: int) -> None:
        assert self.context.session.probe
        self.context.session.probe.set_clock(val)

    def dap_swj_sequence(self, cnt: int, val: int) -> None:
        probe = self.context.session.probe
        assert probe
        if DebugProbe.Capability.SWJ_SEQUENCE not in probe.capabilities:
            raise DebugSequenceRuntimeError(
                    "debug sequence called DAP_SWJ_Sequence, but debug probe does not support this operation")
        probe.swj_sequence(cnt, val)

    def dap_jtag_sequence(self, cnt: int, tms: int, tdi: int) -> int:
        probe = self.context.session.probe
        assert probe
        if DebugProbe.Capability.JTAG_SEQUENCE not in probe.capabilities:
            raise DebugSequenceRuntimeError(
                "debug sequence called DAP_JTAG_Sequence, but debug probe does not support this operation")
        tdo = probe.jtag_sequence(cnt, tms, True, tdi)
        return tdo or 0

    def query(self, type: int, message: str, default: int) -> int:
        LOG.info(f"Query({type}): {message} [{default}]")
        # Just return the default value since we're running in "batch" mode.
        return default

    def queryvalue(self, message: str, default: int) -> int:
        LOG.info(f"QueryValue: {message} [{default}]")
        # Just return the default value since we're running in "batch" mode.
        return default

    _MESSAGE_LEVEL_MAP = {
        0: logging.INFO,
        1: logging.WARNING,
        2: logging.ERROR,
    }
    def message(self, type: int, format: str, *args: int) -> None:
        level = self._MESSAGE_LEVEL_MAP.get(type, 2) # default to error for invalid type
        LOG.log(level, format % args)

    def loaddebuginfo(self, file: str) -> int:
        # Return 1 to indicate failure.
        return 1

# pylint: enable=invalid_name



# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from copy import copy
from typing import (Dict, List, TYPE_CHECKING, Iterable, MutableSequence, Optional, Sequence, Tuple)

from .provider import Breakpoint
from ...core.target import Target

if TYPE_CHECKING:
    from .provider import BreakpointProvider
    from ...core.core_target import CoreTarget
    from ...utility.notification import Notification

LOG = logging.getLogger(__name__)

class UnrealizedBreakpoint(Breakpoint):
    """@brief Breakpoint class used until a breakpoint's type is decided."""
    pass

class BreakpointManager:
    """@brief Manages all breakpoints for one core.

    The most important function of the breakpoint manager is to decide which breakpoint provider
    to use when a breakpoint is added. The caller can request a particular breakpoint type, but
    the manager may decide to use another depending on the situation. For instance, it tries to
    keep one hardware breakpoint available to use for stepping.

    The manager is also responsible for optimising breakpoint adding and removing. When the caller
    requests to add or remove breakpoints, the target is not immediately modified. Instead, the
    add/remove request is recorded for later. Then, before the target is stepped or resumed, the
    manager flushes breakpoint changes to the target. It is at this point when it decides which
    provider to use for each new breakpoint.
    """

    ## Number of hardware breakpoints to try to keep available.
    MIN_HW_BREAKPOINTS = 0

    def __init__(self, core: "CoreTarget") -> None:
        self._breakpoints: Dict[int, Breakpoint] = {}
        self._updated_breakpoints: Dict[int, Breakpoint] = {}
        self._session = core.session
        self._core = core
        self._fpb: Optional["BreakpointProvider"] = None
        self._providers: Dict[Target.BreakpointType, "BreakpointProvider"] = {}
        self._ignore_notifications: bool = False

        # Subscribe to some notifications.
        self._session.subscribe(self._pre_run_handler, Target.Event.PRE_RUN)
        self._session.subscribe(self._pre_disconnect_handler, Target.Event.PRE_DISCONNECT)

    def add_provider(self, provider: "BreakpointProvider") -> None:
        self._providers[provider.bp_type] = provider
        if provider.bp_type == Target.BreakpointType.HW:
            self._fpb = provider

    def get_breakpoints(self) -> Iterable[int]:
        """@brief Return a list of all breakpoint addresses."""
        return self._breakpoints.keys()

    def find_breakpoint(self, addr: int) -> Optional[Breakpoint]:
        return self._updated_breakpoints.get(addr, None)

    def set_breakpoint(self, addr, type=Target.BreakpointType.AUTO):
        """@brief Set a hardware or software breakpoint at a specific location in memory.

        @retval True Breakpoint was set.
        @retval False Breakpoint could not be set.
        """
        LOG.debug("set bkpt type %s at 0x%x", type.name, addr)

        # Clear Thumb bit in case it is set.
        addr = addr & ~1

        # Check for an existing breakpoint at this address.
        bp = self.find_breakpoint(addr)
        if bp is not None:
            return True

        # Reuse breakpoint objects from the live list.
        if addr in self._breakpoints:
            bp = self._breakpoints[addr]
        else:
            # Create temp bp object. This will be replaced with the real object once
            # breakpoints are flushed and the provider sets the bp.
            bp = UnrealizedBreakpoint(self)
            bp.type = type
            bp.addr = addr

            # Check whether this breakpoint can be added when we flush.
            if not self._check_added_breakpoint(bp):
                return False

        self._updated_breakpoints[addr] = bp
        return True

    def _check_added_breakpoint(self, bp: Breakpoint) -> bool:
        """@brief Check whether a new breakpoint is likely to actually be added when we flush.

        First, software breakpoints are assumed to always be addable. For hardware breakpoints,
        the current free hardware breakpoint count is updated based on the current set of to-be
        added and removed breakpoints. If there are enough free hardware breakpoints to meet the
        minimum requirement and still add the new breakpoint, True is returned.
        """
        # If there is no FPB available, just go by whether we can install a sw bp.
        if self._fpb is None:
            region = self._core.memory_map.get_region_for_address(bp.addr)
            return region is not None and region.is_writable

        likely_bp_type = self._select_breakpoint_type(bp, False)
        if likely_bp_type == Target.BreakpointType.SW:
            return True

        # Count updated hw breakpoints.
        free_hw_bp_count = self._fpb.available_breakpoints
        added, removed = self._get_updated_breakpoints()
        for bp in removed:
            if bp.type == Target.BreakpointType.HW:
                free_hw_bp_count += 1
        for bp in added:
            likely_bp_type = self._select_breakpoint_type(bp, False)
            if likely_bp_type == Target.BreakpointType.HW:
                free_hw_bp_count -= 1

        return free_hw_bp_count > self.MIN_HW_BREAKPOINTS

    def remove_breakpoint(self, addr: int) -> None:
        """@brief Remove a breakpoint at a specific location."""
        try:
            LOG.debug("remove bkpt at 0x%x", addr)

            # Clear Thumb bit in case it is set.
            addr = addr & ~1

            # Remove bp from dict.
            del self._updated_breakpoints[addr]
        except KeyError:
            LOG.debug("Tried to remove breakpoint 0x%08x that wasn't set" % addr)

    def _get_updated_breakpoints(self) -> Tuple[List[Breakpoint], List[Breakpoint]]:
        """@brief Compute added and removed breakpoints since last flush.
        @return Bi-tuple of (added breakpoint list, removed breakpoint list).
        """
        added = []
        removed = []

        # Get added breakpoints.
        for bp in self._updated_breakpoints.values():
            if not bp.addr in self._breakpoints:
                added.append(bp)

        # Get removed breakpoints.
        for bp in self._breakpoints.values():
            if not bp.addr in self._updated_breakpoints:
                removed.append(bp)

        # Return the list of pages to update.
        return added, removed

    def _select_breakpoint_type(self, bp: Breakpoint, allow_all_hw_bps: bool) -> Optional[Target.BreakpointType]:
        type = bp.type

        # Look up the memory type for the requested address.
        region = self._core.memory_map.get_region_for_address(bp.addr)
        if region is not None:
            is_writable = region.is_writable
        else:
            # No memory region - fallback to hardware breakpoints.
            type = Target.BreakpointType.HW
            is_writable = False

        in_hw_bkpt_range = (self._fpb is not None) and (self._fpb.can_support_address(bp.addr))
        have_hw_bp = (self._fpb is not None) \
                    and ((self._fpb.available_breakpoints > self.MIN_HW_BREAKPOINTS) \
                    or (allow_all_hw_bps and self._fpb.available_breakpoints > 0))

        # Determine best type to use if auto.
        if type == Target.BreakpointType.AUTO:
            # Use sw breaks for:
            #  1. Addresses outside the supported FPBv1 range of 0-0x1fffffff
            #  2. RAM regions by default.
            #  3. Number of remaining hw breaks are at or less than the minimum we want to keep.
            #
            # Otherwise use hw.
            if not in_hw_bkpt_range or (not have_hw_bp):
                if is_writable:
                    type = Target.BreakpointType.SW
                else:
                    LOG.debug("unable to set bp because no hw bp available")
                    return None
            else:
                type = Target.BreakpointType.HW

            LOG.debug("using type %s for auto bp", type.name)

        # Revert to sw bp if out of hardware breakpoint range.
        if (type == Target.BreakpointType.HW) and not in_hw_bkpt_range:
            if is_writable:
                LOG.debug("using sw bp instead because of unsupported addr")
                type = Target.BreakpointType.SW
            else:
                LOG.debug("could not fallback to software breakpoint")
                return None

        # Revert to hw bp if region is flash.
        if not is_writable:
            if in_hw_bkpt_range and have_hw_bp:
                LOG.debug("using hw bp instead because addr is flash")
                type = Target.BreakpointType.HW
            else:
                LOG.debug("could not fallback to hardware breakpoint")
                return None

        LOG.debug("selected bkpt type %s for addr 0x%x", type.name, bp.addr)
        return type

    def flush(self, is_step: bool = False) -> None:
        try:
            # Ignore any notifications while we modify breakpoints.
            self._ignore_notifications = True

            added, removed = self._get_updated_breakpoints()
            LOG.debug("added=%s removed=%s", added, removed)

            # Handle removed breakpoints first by asking the providers to remove them.
            for bp in removed:
                assert bp.provider is not None
                bp.provider.remove_breakpoint(bp)
                del self._breakpoints[bp.addr]

            # Only allow use of all hardware breakpoints if we're not stepping and there is
            # only a single added breakpoint.
            allow_all_hw_bps = not is_step and len(added) == 1

            # Now handle added breakpoints.
            for bp in added:
                type = self._select_breakpoint_type(bp, allow_all_hw_bps)
                if type is None:
                    continue

                # Set the bp.
                try:
                    provider = self._providers[type]
                    bp = provider.set_breakpoint(bp.addr)
                except KeyError:
                    raise ValueError("Unknown breakpoint type %s" % type.name)

                # Save the bp.
                if bp is not None:
                    self._breakpoints[bp.addr] = bp

            # Update breakpoint lists.
            LOG.debug("bps after flush=%s", self._breakpoints)
            self._updated_breakpoints = copy(self._breakpoints)

            # Flush all providers.
            self._flush_all()
        finally:
            self._ignore_notifications = False

    def get_breakpoint_type(self, addr: int) -> Optional[Target.BreakpointType]:
        bp = self.find_breakpoint(addr)
        return bp.type if (bp is not None) else None

    def filter_memory(self, addr: int, size: int, data: int) -> int:
        for provider in [p for p in self._providers.values() if p.do_filter_memory]:
            data = provider.filter_memory(addr, size, data)
        return data

    def filter_memory_unaligned_8(self, addr: int, size: int, data: MutableSequence[int]) -> Sequence[int]:
        for provider in [p for p in self._providers.values() if p.do_filter_memory]:
            for i, d in enumerate(data):
                data[i] = provider.filter_memory(addr + i, 8, d)
        return data

    def filter_memory_aligned_32(self, addr: int, size: int, data: MutableSequence[int]) -> Sequence[int]:
        for provider in [p for p in self._providers.values() if p.do_filter_memory]:
            for i, d in enumerate(data):
                data[i] = provider.filter_memory(addr + i, 32, d)
        return data

    def remove_all_breakpoints(self) -> None:
        """@brief Remove all breakpoints immediately."""
        for bp in self._breakpoints.values():
            bp.provider.remove_breakpoint(bp)
        self._breakpoints = {}
        self._flush_all()

    def _flush_all(self) -> None:
        # Flush all providers.
        for provider in self._providers.values():
            provider.flush()

    def _pre_run_handler(self, notification: "Notification") -> None:
        if not self._ignore_notifications:
            is_step = notification.data == Target.RunType.STEP
            self.flush(is_step)

    def _pre_disconnect_handler(self, notification: "Notification") -> None:
        pass





# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from enum import (Enum, IntFlag)
import threading
from typing import (Callable, Collection, Optional, overload, Sequence, Set, TYPE_CHECKING, Tuple, Union)
from typing_extensions import Literal

if TYPE_CHECKING:
    from ..core.session import Session
    from ..core.memory_interface import MemoryInterface
    from ..board.board import Board
    from ..board.board_ids import BoardInfo
    from ..coresight.ap import APAddressBase

class DebugProbe:
    """@brief Abstract debug probe class.

    Subclasses of this abstract class are drivers for different debug probe interfaces, either hardware such as a
    USB based probe, or software such as connecting with a simulator.

    The constructor is private. To create an instance, use either of get_all_connected_probes() or get_probe_with_id().
    Normally, the @ref pyocd.probe.aggregator.DebugProbeAggregator "DebugProbeAggregator" class is used instead of
    directly calling methods on a specific probe class.

    Use an instance as follows:

    1. Call open().
    2. Optionally inspect the `supported_wire_protocols` property and select a protocol to use.
    3. Call connect(), passing the chosen wire protocol.
    4. Use by instance by calling other methods.
    5. Call disconnect().
    6. Call close().

    Most methods are required to be overridden by a subclass, with a few exceptions.

    These methods are completely optional:

    - create_associated_board()
    - flush()
    - get_memory_interface_for_ap()

    These methods must be implemented depending on the probe capabilities, as returned from the `capabilities` property.

    - swj_sequence(): Capability.SWJ_SEQUENCE; if not provided it is assumed the probe automatically enables SWD or JTAG
        on the target based on the protocol passed into connect().
    - swd_sequence(): Capability.SWD_SEQUENCE
    - jtag_sequence(): Capability.JTAG_SEQUENCE
    - swo_*(): Capability.SWO
    - get_accessible_pins(): Capability.PIN_ACCESS
    - read_pins(): Capability.PIN_ACCESS
    - write_pins(): Capability.PIN_ACCESS
    """

    class Protocol(Enum):
        """@brief Debug wire protocols."""
        DEFAULT = 0
        SWD = 1
        JTAG = 2

    class PinGroup(Enum):
        """@brief Available pin groups for read/write pins APIs."""
        PROTOCOL_PINS = 0
        GPIO_PINS = 1

    class ProtocolPin(IntFlag):
        """@brief Pin mask constants for SWD/JTAG protocol pins."""
        SWCLK_TCK = 1 << 0
        SWDIO_TMS = 1 << 1
        TDI = 1 << 2
        TDO = 1 << 3
        nRESET = 1 << 4
        nTRST = 1 << 5
        ALL_PINS = SWCLK_TCK | SWDIO_TMS | TDI | TDO | nRESET | nTRST

    ## Map from wire protocol setting name to debug probe constant.
    PROTOCOL_NAME_MAP = {
            'swd': Protocol.SWD,
            'jtag': Protocol.JTAG,
            'default': Protocol.DEFAULT,
        }

    class Capability(Enum):
        """@brief Probe capabilities."""
        ## @brief Whether the probe supports the swj_sequence() API.
        #
        # If this property is True, then the swj_sequence() method is used to move between protocols.
        # If False, it is assumed the probe firmware automatically manages the protocol switch.
        SWJ_SEQUENCE = 0

        ## @brief Whether the probe supports receiving SWO data.
        SWO = 1

        ## @brief Whether the probe can access banked DP registers.
        #
        # Currently only used to verify that the probe supports banked DP registers when the #MANAGED_DPBANKSEL
        # capability is present.
        BANKED_DP_REGISTERS = 2

        ## @brief Whether the probe can access APv2 registers.
        #
        # This capability is currently only used to verify that a probe with the #MANAGED_AP_SELECTION capability
        # can support the wider AP addresses used in version 2 APs. For probes without #MANAGED_AP_SELECTION,
        # DP_SELECT is written directly by the DAP layer when selecting an AP.
        APv2_ADDRESSES = 3

        ## @brief Whether the probe automatically handles AP selection in the DP.
        #
        # If this capability is not present, the DebugPort object will perform the AP selection
        # by DP register writes.
        MANAGED_AP_SELECTION = 4

        ## @brief whether the probe automatically handles access of banked DAP registers.
        MANAGED_DPBANKSEL = 5

        ## @brief Whether the probe supports the swd_sequence() API.
        SWD_SEQUENCE = 6

        ## @brief Whether the probe supports the jtag_sequence() API.
        JTAG_SEQUENCE = 7

        ## @brief Pin access via the read_pins()/write_pins() APIs.
        PIN_ACCESS = 8

    @classmethod
    def get_all_connected_probes(
                cls,
                unique_id: Optional[str] = None,
                is_explicit: bool = False
            ) -> Sequence[DebugProbe]:
        """@brief Returns a list of DebugProbe instances.

        To filter the list of returned probes, the `unique_id` parameter may be set to a string with a full or
        partial unique ID (canonically the serial number). Alternatively, the probe class may simply return all
        available probes and let the caller handle filtering.

        @param cls The class instance.
        @param unique_id String. Optional partial unique ID value used to filter available probes. May be used by the
            probe to optimize retrieving the probe list; there is no requirement to filter the results.
        @param is_explicit Boolean. Whether the probe type was explicitly specified in the unique ID. This
            can be used, for instance, to specially interpret the unique ID as an IP address or
            domain name when the probe class was specifically requested but not for general lists
            of available probes.
        @return List of DebugProbe instances.
        """
        raise NotImplementedError()

    @classmethod
    def get_probe_with_id(cls, unique_id: str, is_explicit: bool = False) -> Optional[DebugProbe]:
        """@brief Returns a DebugProbe instance for a probe with the given unique ID.

        If no probe is connected with a fully matching unique ID, then None will be returned.

        @param cls The class instance.
        @param unique_id Unique ID string to match against probes' full unique ID. No partial matches are allowed.
        @param is_explicit Boolean. Whether the probe type was explicitly specified in the unique ID.
        @return DebugProbe instance, or None
        """
        raise NotImplementedError()

    def __init__(self) -> None:
        """@brief Constructor."""
        self._session: Optional[Session] = None
        self._lock = threading.RLock()

    @property
    def session(self) -> Optional[Session]:
        """@brief Session associated with this probe."""
        return self._session

    @session.setter
    def session(self, the_session: Session) -> None:
        self._session = the_session

    @property
    def description(self) -> str:
        """@brief Combined description of the debug probe and/or associated board."""
        return self.vendor_name + " " + self.product_name

    @property
    def vendor_name(self) -> str:
        """@brief Name of the debug probe's manufacturer."""
        raise NotImplementedError()

    @property
    def product_name(self) -> str:
        """@brief Name of the debug probe."""
        raise NotImplementedError()

    @property
    def supported_wire_protocols(self) -> Collection[Protocol]:
        """@brief List of DebugProbe.Protocol supported by the probe.

        Only one of the values returned from this property may be passed to connect().
        """
        raise NotImplementedError()

    @property
    def unique_id(self) -> str:
        """@brief The unique ID of this device.

        This property will be valid before open() is called. This value can be passed to
        get_probe_with_id().
        """
        raise NotImplementedError()

    @property
    def wire_protocol(self) -> Optional[Protocol]:
        """@brief Currently selected wire protocol.

        If the probe is not open and connected, i.e., open() and connect() have not been called,
        then this property will be None. If a value other than None is returned, then the probe
        has been connected successfully.
        """
        raise NotImplementedError()

    @property
    def is_open(self) -> bool:
        """@brief Whether the probe is currently open.

        To open the probe, call the open() method.
        """
        raise NotImplementedError()

    @property
    def capabilities(self) -> Set[Capability]:
        """@brief A set of DebugProbe.Capability enums indicating the probe's features.

        This value should not be trusted until after the probe is opened.
        """
        raise NotImplementedError()

    @property
    def associated_board_info(self) -> Optional[BoardInfo]:
        """@brief Info about the board associated with this probe, if known."""
        return None

    def create_associated_board(self) -> Optional[Board]:
        """@brief Create a board instance representing the board of which the probe is a component.

        If the probe is part of a board, then this method will create a Board instance that
        represents the associated board. Usually, for an on-board debug probe, this would be the
        Board that the probe physically is part of, and will also set the target type. If the probe
        does not have an associated board, then this method returns None.

        @param self
        """
        return None

    def get_accessible_pins(self, group: PinGroup) -> Tuple[int, int]:
        """@brief Return masks of pins accessible via the .read_pins()/.write_pins() methods.

        This method is only expected to be implemented if Capability.PIN_ACCESS is present.

        @return Tuple of pin masks for (0) readable, (1) writable pins. See DebugProbe.Pin for mask
        values for those pins that have constants.
        """
        raise NotImplementedError()

    def open(self) -> None:
        """@brief Open the USB interface to the probe for sending commands."""
        raise NotImplementedError()

    def close(self) -> None:
        """@brief Close the probe's USB interface."""
        raise NotImplementedError()

    def lock(self) -> None:
        """@brief Lock the probe from access by other threads.

        This lock is recursive, so locking multiple times from a single thread is acceptable as long
        as the thread unlocks the same number of times.

        This method does not return until the calling thread has ownership of the lock.
        """
        self._lock.acquire()

    def unlock(self) -> None:
        """@brief Unlock the probe.

        Only when the thread unlocks the probe the same number of times it has called lock() will
        the lock actually be released and other threads allowed access.
        """
        self._lock.release()

    ## @name Target control
    ##@{

    def connect(self, protocol: Optional[Protocol] = None) -> None:
        """@brief Initialize DAP IO pins for JTAG or SWD"""
        raise NotImplementedError()

    def disconnect(self) -> None:
        """@brief Deinitialize the DAP I/O pins"""
        raise NotImplementedError()

    def swj_sequence(self, length: int, bits: int) -> None:
        """@brief Transfer some number of bits on SWDIO/TMS.

        @param self
        @param length Number of bits to transfer. Must be less than or equal to 256.
        @param bits Integer of the bit values to send on SWDIO/TMS. The LSB is transmitted first.
        """
        pass

    def swd_sequence(self, sequences: Sequence[Union[Tuple[int], Tuple[int, int]]]) -> Tuple[int, Sequence[bytes]]:
        """@brief Send a sequences of bits on the SWDIO signal.

        Each sequence in the _sequences_ parameter is a tuple with 1 or 2 members in this order:
        - 0: int: number of TCK cycles from 1-64
        - 1: int: the SWDIO bit values to transfer. The presence of this tuple member indicates the sequence is
            an output sequence; the absence means that the specified number of TCK cycles of SWDIO data will be
            read and returned.

        @param self
        @param sequences A sequence of sequence description tuples as described above.

        @return A 2-tuple of the response status, and a sequence of bytes objects, one for each input
            sequence. The length of the bytes object is (<TCK-count> + 7) / 8. Bits are in LSB first order.
        """
        raise NotImplementedError()

    def jtag_sequence(self, cycles: int, tms: int, read_tdo: bool, tdi: int) -> Optional[int]:
        """@brief Send JTAG sequence.

        @param self
        @param cycles Number of TCK cycles, from 1-64.
        @param tms Fixed TMS value. Either 0 or 1.
        @param read_tdo Boolean indicating whether TDO should be read.
        @param tdi Integer with the TDI bit values to be transferred each TCK cycle. The LSB is
            sent first.

        @return Either an integer with TDI bit values, or None, if _read_tdo_ was false.
        """
        raise NotImplementedError()

    def set_clock(self, frequency: float) -> None:
        """@brief Set the frequency for JTAG and SWD in Hz.

        This function is safe to call before connect is called.
        """
        raise NotImplementedError()

    def reset(self) -> None:
        """@brief Perform a hardware reset of the target."""
        raise NotImplementedError()

    def assert_reset(self, asserted: bool) -> None:
        """@brief Assert or de-assert target's nRESET signal.

        Because nRESET is negative logic and usually open drain, passing True will drive it low, and
        passing False will stop driving so nRESET will be pulled up.
        """
        raise NotImplementedError()

    def is_reset_asserted(self) -> bool:
        """@brief Returns True if nRESET is asserted or False if de-asserted.

        If the debug probe cannot actively read the reset signal, the value returned will be the
        last value passed to assert_reset().
        """
        raise NotImplementedError()

    def flush(self) -> None:
        """@brief Write out all unsent commands.

        This API may be a no-op for certain debug probe types.
        """
        pass

    def read_pins(self, group: PinGroup, mask: int) -> int:
        """@brief Read values of selected debug probe pins.

        See DebugProbe.ProtocolPin for mask values for the DebugProbe.PinGroup.PROTOCOL_PINS group.

        This method is only expected to be implemented if Capability.PIN_ACCESS is present.

        @param self
        @param group Select the pin group to read.
        @param mask Bit mask indicating which pins will be read. The return value will contain only
            bits set in this mask.
        @return Bit mask with the current value of selected pins at each pin's relevant bit position.
        """
        raise NotImplementedError()

    def write_pins(self, group: PinGroup, mask: int, value: int) -> None:
        """@brief Set values of selected debug probe pins.

        See DebugProbe.ProtocolPin for mask values for the DebugProbe.PinGroup.PROTOCOL_PINS group.
        Note that input-only pins such as TDO are not writable with most debug probes.

        This method is only expected to be implemented if Capability.PIN_ACCESS is present.

        @param self
        @param group Select the pin group to read.
        @param mask Bit mask indicating which pins will be written.
        @param value Mask containing the bit value of to written for selected pins at each pin's
            relevant bit position..
        """
        raise NotImplementedError()

    ##@}

    ## @name DAP access
    ##@{

    @overload
    def read_dp(self, addr: int) -> int:
        ...

    @overload
    def read_dp(self, addr: int, now: Literal[True] = True) -> int:
        ...

    @overload
    def read_dp(self, addr: int, now: Literal[False]) -> Callable[[], int]:
        ...

    @overload
    def read_dp(self, addr: int, now: bool) -> Union[int, Callable[[], int]]:
        ...

    def read_dp(self, addr: int, now: bool = True) -> Union[int, Callable[[], int]]:
        """@brief Read a DP register.

        @param self
        @param addr Integer register address being one of (0x0, 0x4, 0x8, 0xC).
        @param now Boolean specifying whether the read is synchronous (True) or asynchronous.
        @return If _now_ is True, the register's 32-bit value is returned as an integer. When _now_
            is False, a callable is returned that when invoked will return the register's value.
        """
        raise NotImplementedError()

    def write_dp(self, addr: int, data: int) -> None:
        """@brief Write a DP register.

        @param self
        @param addr Integer register address being one of (0x0, 0x4, 0x8, 0xC).
        @param data Integer register value.
        """
        raise NotImplementedError()

    @overload
    def read_ap(self, addr: int) -> int:
        ...

    @overload
    def read_ap(self, addr: int, now: Literal[True] = True) -> int:
        ...

    @overload
    def read_ap(self, addr: int, now: Literal[False]) -> Callable[[], int]:
        ...

    @overload
    def read_ap(self, addr: int, now: bool) -> Union[int, Callable[[], int]]:
        ...

    def read_ap(self, addr: int, now: bool = True) -> Union[int, Callable[[], int]]:
        """@brief Read an AP register."""
        raise NotImplementedError()

    def write_ap(self, addr: int, data) -> None:
        """@brief Write an AP register."""
        raise NotImplementedError()

    @overload
    def read_ap_multiple(self, addr: int, count: int = 1) -> Sequence[int]:
        ...

    @overload
    def read_ap_multiple(self, addr: int, count: int, now: Literal[True] = True) -> Sequence[int]:
        ...

    @overload
    def read_ap_multiple(self, addr: int, count: int, now: Literal[False]) -> Callable[[], Sequence[int]]:
        ...

    @overload
    def read_ap_multiple(self, addr: int, count: int, now: bool) -> Union[Sequence[int], Callable[[], Sequence[int]]]:
        ...

    def read_ap_multiple(self, addr: int, count: int = 1, now: bool = True) \
             -> Union[Sequence[int], Callable[[], Sequence[int]]]:
        """@brief Read one AP register multiple times."""
        raise NotImplementedError()

    def write_ap_multiple(self, addr: int, values) -> None:
        """@brief Write one AP register multiple times."""
        raise NotImplementedError()

    def get_memory_interface_for_ap(self, ap_address: APAddressBase) -> Optional[MemoryInterface]:
        """@brief Returns a @ref pyocd.core.memory_interface.MemoryInterface "MemoryInterface" for
            the specified AP.

        Some debug probe types have accelerated memory read and write commands. This method is used
        to get a concrete @ref pyocd.core.memory_interface.MemoryInterface "MemoryInterface"
        instance that is specific to the AP identified by the _ap_address_ parameter. If the probe
        does not provide an accelerated memory interface, None will be returned.

        @param self The debug probe.
        @param ap_address An instance of @ref pyocd.coresight.ap.APAddressBase "APAddressBase".
        """
        return None

    ##@}

    ## @name SWO
    ##@{

    def swo_start(self, baudrate: float) -> None:
        """@brief Start receiving SWO data at the given baudrate.

        Once SWO reception has started, the swo_read() method must be called at regular intervals
        to receive SWO data. If this is not done, the probe's internal SWO data buffer may overflow
        and data will be lost.
        """
        raise NotImplementedError()

    def swo_stop(self) -> None:
        """@brief Stop receiving SWO data."""
        raise NotImplementedError()

    def swo_read(self) -> bytearray:
        """@brief Read buffered SWO data from the target.

        @eturn Bytearray of the received data. May be 0 bytes in length if no SWO data is buffered
            at the probe.
        """
        raise NotImplementedError()

    ##@}

    def __repr__(self):
        return "<{}@{:x} {}>".format(self.__class__.__name__, id(self), self.description)





# pyOCD debugger
# Copyright (c) 2016 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

LOG = logging.getLogger(__name__)

class TargetThread(object):
    """@brief Base class representing a thread on the target."""

    def __init__(self):
        pass

    @property
    def unique_id(self):
        raise NotImplementedError()

    @property
    def name(self):
        raise NotImplementedError()

    @property
    def description(self):
        raise NotImplementedError()

    @property
    def is_current(self):
        raise NotImplementedError()

    @property
    def context(self):
        raise NotImplementedError()

class ThreadProvider(object):
    """@brief Base class for RTOS support plugins."""

    def __init__(self, target):
        self._target = target
        self._target_context = self._target.get_target_context()
        self._last_run_token = -1
        self._read_from_target = False

    def _lookup_symbols(self, symbolList, symbolProvider):
        syms = {}
        for name in symbolList:
            addr = symbolProvider.get_symbol_value(name)
            LOG.debug("Value for symbol %s = %s", name, hex(addr) if addr is not None else "<none>")
            if addr is None:
                return None
            syms[name] = addr
        return syms

    def init(self, symbolProvider):
        """@retval True The provider was successfully initialzed.
        @retval False The provider could not be initialized successfully.
        """
        raise NotImplementedError()

    def _build_thread_list(self):
        raise NotImplementedError()

    def _is_thread_list_dirty(self):
        token = self._target.run_token
        if token == self._last_run_token:
            # Target hasn't run since we last updated threads, so there is nothing to do.
            return False
        self._last_run_token = token
        return True

    def update_threads(self):
        if self._is_thread_list_dirty() and self._read_from_target:
            self._build_thread_list()

    def get_threads(self):
        raise NotImplementedError()

    def get_thread(self, threadId):
        raise NotImplementedError()

    def invalidate(self):
        raise NotImplementedError()

    @property
    def read_from_target(self):
        return self._read_from_target

    @read_from_target.setter
    def read_from_target(self, value):
        if value != self._read_from_target:
            self.invalidate()
        self._read_from_target = value

    @property
    def is_enabled(self):
        raise NotImplementedError()

    @property
    def current_thread(self):
        raise NotImplementedError()

    def is_valid_thread_id(self, threadId):
        raise NotImplementedError()

    def get_current_thread_id(self):
        """From GDB's point of view, where Handler Mode is a thread"""
        raise NotImplementedError()

    def get_actual_current_thread_id(self):
        """From OS's point of view, so the current OS thread even in Handler Mode"""
        raise NotImplementedError()



# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import colorama
from colorama import (Fore, Style)
import logging
from shutil import get_terminal_size
import sys
from typing import (IO, Optional)

class ColorFormatter(logging.Formatter):
    """@brief Log formatter that applies colours based on the record's log level."""

    FORMAT = "{timecolor}{relativeCreated:07.0f}{_reset} {lvlcolor:s}{levelname:<{levelnamewidth}.{levelnamewidth}s}{_reset} {msgcolor}{message} {_dim}[{module:s}]{_reset}"

    ## Colors for the log level name.
    LEVEL_COLORS = {
            'CRITICAL': Style.BRIGHT + Fore.LIGHTRED_EX,
            'ERROR': Fore.LIGHTRED_EX,
            'WARNING': Fore.LIGHTYELLOW_EX,
            'INFO': Fore.CYAN,
            'DEBUG': Style.DIM,
        }

    ## Colors for the rest of the log message.
    MESSAGE_COLORS = {
            'CRITICAL': Fore.LIGHTRED_EX,
            'ERROR': Fore.RED,
            'WARNING': Fore.YELLOW,
            'DEBUG': Style.DIM + Fore.LIGHTWHITE_EX,
        }

    ## Fixed maximum length of the log level name in log messages.
    MAX_LEVELNAME_WIDTH = 1

    def __init__(self, msg, use_color: bool, is_tty: bool) -> None:
        super().__init__(msg, style='{')
        self._use_color = use_color
        self._is_tty = is_tty

        # TODO: Handle resizing of terminal?
        self._term_width = get_terminal_size()[0]

    # Note: we can't set the type of `record` param to LogRecord because that causes type errors for
    # each time below when an attribute is set on the record.
    def format(self, record) -> str:
        # Capture and remove exc_info and stack_info so the superclass format() doesn't
        # print it and we can control the formatting.
        exc_info = record.exc_info
        record.exc_info = None
        stack_info = record.stack_info
        record.stack_info = None

        # Add colors to the record.
        if self._use_color:
            record.lvlcolor = self.LEVEL_COLORS.get(record.levelname, '')

            # Colorise the line.
            record.msgcolor = self.MESSAGE_COLORS.get(record.levelname, '')

            # Fixed colors.
            record.timecolor = Fore.BLUE
            record._reset = Style.RESET_ALL
            record._dim = Style.DIM
        else:
            record.lvlcolor = ""
            record.msgcolor = ""
            record.timecolor = ""
            record._reset = ""
            record._dim = ""

        record.message = record.getMessage()

        # Add levelname alignment to record.
        record.levelname_align = " " * max(self.MAX_LEVELNAME_WIDTH - len(record.levelname), 0)
        record.levelnamewidth = self.MAX_LEVELNAME_WIDTH

        # Let superclass handle formatting.
        log_msg = super().format(record)

        # Append uncolored exception/stack info.
        if exc_info:
            log_msg += "\n" + Style.DIM + self.formatException(exc_info) + Style.RESET_ALL
        if stack_info:
            log_msg += "\n" + Style.DIM + self.formatStack(stack_info) + Style.RESET_ALL

        return log_msg


def build_color_logger(
            level: int = logging.INFO,
            color_setting: str = 'auto',
            stream: Optional[IO[str]] = None,
            is_tty: Optional[bool] = None,
        ) -> logging.Logger:
    """@brief Sets up color logging for the root logger.

    @param level Log level of the root logger.
    @param color_setting One of 'auto', 'always', or 'never'. The default 'auto' enables color if `is_tty` is True.
    @param stream The stream to which the log will be output. The default is stderr.
    @param is_tty Whether the output stream is a tty. Affects the 'auto' color_setting. If not provided, the
        `isatty()` method of both `sys.stdout` and `sys.stderr` is called if they exists. Both must return
        True for the 'auto' color setting to enable color output.
    """
    if stream is None:
        stream = sys.stderr
    if is_tty is None:
        stdout_is_tty = sys.stdout.isatty() if hasattr(sys.stdout, 'isatty') else False
        stderr_is_tty = sys.stderr.isatty() if hasattr(sys.stderr, 'isatty') else False
        is_tty = stdout_is_tty and stderr_is_tty
    use_color = (color_setting == "always") or (color_setting == "auto" and is_tty)

    # Init colorama with appropriate color setting.
    colorama.init(strip=(not use_color))

    # Create handler to output logging to stderr.
    console = logging.StreamHandler(stream)

    # Create the color formatter and attach to our stream handler.
    color_formatter = ColorFormatter(ColorFormatter.FORMAT, use_color, is_tty)
    console.setFormatter(color_formatter)

    # Set stream handler and log level on root logger.
    root_logger = logging.getLogger()
    root_logger.addHandler(console)
    root_logger.setLevel(level)

    return root_logger




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import (Callable, Dict, List, Optional, overload, Sequence, Union, TYPE_CHECKING)
from typing_extensions import Literal

from .target import (Target, TargetGraphNode)
from .core_target import CoreTarget
from ..flash.eraser import FlashEraser
from ..debug.cache import CachingDebugContext
from ..debug.context import DebugContext
from ..debug.elf.elf import ELFBinaryFile
from ..debug.elf.elf_reader import ElfReaderContext
from ..utility.sequencer import CallSequence

if TYPE_CHECKING:
    from .session import Session
    from .memory_map import MemoryMap
    from .core_registers import (CoreRegistersIndex, CoreRegisterNameOrNumberType, CoreRegisterValueType)
    from ..debug.context import DebugContext
    from ..debug.breakpoints.provider import Breakpoint
    from ..commands.execution_context import CommandSet

LOG = logging.getLogger(__name__)

class SoCTarget(TargetGraphNode):
    """@brief Represents a microcontroller system-on-chip.

    An instance of this class is the root of the chip-level object graph. It has child
    nodes for the DP and all cores. As a concrete subclass of Target, it provides methods
    to control the device, access memory, adjust breakpoints, and so on.

    For single core devices, the SoCTarget has mostly equivalent functionality to
    the CoreTarget object for the core. Multicore devices work differently. This class tracks
    a "selected core", to which all actions are directed. The selected core can be changed
    at any time. You may also directly access specific cores and perform operations on them.

    SoCTarget subclasses must restrict usage of the DebugProbe instance in their constructor, ideally not
    using it at all. This is required in order to be able to gather information about targets for commands
    such as `pyocd json` and `pyocd list`. These commands create a session with the probe set to an instance
    of StubProbe, which is a subclass of DebugProbe with the minimal implementation necessary to support
    session creation but not opening.
    """

    VENDOR = "Generic"

    def __init__(self, session: Session, memory_map: Optional[MemoryMap] = None) -> None:
        super().__init__(session, memory_map)
        self.vendor: str = self.VENDOR
        self.part_families: List[str] = getattr(self, 'PART_FAMILIES', [])
        self.part_number: str = getattr(self, 'PART_NUMBER', self.__class__.__name__)
        self._cores: Dict[int, CoreTarget] = {}
        self._selected_core: int = -1
        self._new_core_num = 0
        self._elf = None

        # Set our graph node name.
        self.node_name = 'soc'

    @property
    def cores(self) -> Dict[int, CoreTarget]:
        return self._cores

    @property
    def selected_core(self) -> Optional[CoreTarget]:
        """@brief Get the selected CPU core object."""
        if self._selected_core == -1:
            return None
        return self.cores[self._selected_core]

    @selected_core.setter
    def selected_core(self, core_number: int) -> None:  # type:ignore # core_number int type is not the same
                                                                      # as selected_core property return type
        """@brief Set the selected CPU core object."""
        if core_number not in self.cores:
            raise ValueError("invalid core number %d" % core_number) # TODO should be a KeyError
        LOG.debug("selected core #%d" % core_number)
        self._selected_core = core_number

    @property
    def selected_core_or_raise(self) -> CoreTarget:
        """@brief Get the selected CPU core object.

        Like selected_core but will raise an exception if no core is selected rather than returning None.
        @exception KeyError The selected_core property is None.
        """
        if self._selected_core == -1:
            raise KeyError("SoCTarget has no selected core")
        return self.cores[self._selected_core]

    @property
    def primary_core(self) -> CoreTarget:
        """@brief Return the core for the `primary_core` session option.

        @exception KeyError The `primary_core` option is invalid.
        """
        primary_core_number = self.session.options.get('primary_core')
        return self.cores[primary_core_number]

    @property
    def elf(self) -> Optional[ELFBinaryFile]:
        return self._elf

    @elf.setter
    def elf(self, filename: str) -> None: # type:ignore # filename str type is not same as elf property return type
        if filename is None:
            self._elf = None
        else:
            self._elf = ELFBinaryFile(filename, self.memory_map)
            for core_number in range(len(self.cores)):
                self.cores[core_number].elf = self._elf
                if self.session.options['cache.read_code_from_elf']:
                    self.cores[core_number].set_target_context(
                            ElfReaderContext(self.cores[core_number].get_target_context(), self._elf))

    @property
    def supported_security_states(self) -> Sequence[Target.SecurityState]:
        return self.selected_core_or_raise.supported_security_states

    @property
    def core_registers(self) -> CoreRegistersIndex:
        return self.selected_core_or_raise.core_registers

    def add_core(self, core: CoreTarget) -> None:
        core.delegate = self.delegate
        if self.debug_sequence_delegate:
            core.debug_sequence_delegate = self.debug_sequence_delegate
        ctx = CachingDebugContext(
                core,
                enable_memory=self.session.options['cache.enable_memory'],
                enable_register=self.session.options['cache.enable_register'],
                )
        core.set_target_context(ctx)
        self.cores[core.core_number] = core
        self.add_child(core)

        # Always select first added core to ensure some core is selected.
        if self.selected_core is None:
            self.selected_core = core.core_number
        # Otherwise, when the chosen primary core is added, select it. This assumes that cores are only
        # added at init/discovery time.
        elif core.core_number == self.session.options.get('primary_core'):
            self.selected_core = core.core_number

    def create_init_sequence(self) -> CallSequence:
        # Return an empty call sequence. The subclass must override this.
        return CallSequence()

    def init(self) -> None:
        # If we don't have a delegate installed yet but there is a session delegate, use it.
        if (self.delegate is None) and (self.session.delegate is not None):
            self.delegate = self.session.delegate

        # Create and execute the init sequence.
        seq = self.create_init_sequence()
        self.call_delegate('will_init_target', target=self, init_sequence=seq)
        seq.invoke()
        self.call_delegate('did_init_target', target=self)

    def post_connect_hook(self) -> None:
        """@brief Hook function called after post_connect init task.

        This hook lets the target subclass configure the target as necessary.
        """
        pass

    def disconnect(self, resume: bool = True) -> None:
        self.session.notify(Target.Event.PRE_DISCONNECT, self)
        self.call_delegate('will_disconnect', target=self, resume=resume)
        for core in self.cores.values():
            core.disconnect(resume)
        self.call_delegate('did_disconnect', target=self, resume=resume)

    @property
    def run_token(self) -> int:
        return self.selected_core_or_raise.run_token

    def halt(self) -> None:
        return self.selected_core_or_raise.halt()

    def step(self, disable_interrupts: bool = True, start: int = 0, end: int = 0,
            hook_cb: Optional[Callable[[], bool]] = None) -> None:
        return self.selected_core_or_raise.step(disable_interrupts, start, end, hook_cb)

    def resume(self) -> None:
        return self.selected_core_or_raise.resume()

    def mass_erase(self) -> None:
        if not self.call_delegate('mass_erase', target=self):
            # The default mass erase implementation is to simply perform a chip erase.
            eraser = FlashEraser(self.session, FlashEraser.Mode.CHIP)
            eraser._log_chip_erase = False
            eraser.erase()

    def write_memory(self, addr: int, data: int, transfer_size: int = 32) -> None:
        return self.selected_core_or_raise.write_memory(addr, data, transfer_size)

    @overload
    def read_memory(self, addr: int, transfer_size: int = 32) -> int:
        ...

    @overload
    def read_memory(self, addr: int, transfer_size: int = 32, now: Literal[True] = True) -> int:
        ...

    @overload
    def read_memory(self, addr: int, transfer_size: int, now: Literal[False]) -> Callable[[], int]:
        ...

    @overload
    def read_memory(self, addr: int, transfer_size: int, now: bool) -> Union[int, Callable[[], int]]:
        ...

    def read_memory(self, addr: int, transfer_size: int = 32, now: bool = True) -> Union[int, Callable[[], int]]:
        return self.selected_core_or_raise.read_memory(addr, transfer_size, now)

    def write_memory_block8(self, addr: int, data: Sequence[int]) -> None:
        return self.selected_core_or_raise.write_memory_block8(addr, data)

    def write_memory_block32(self, addr: int, data: Sequence[int]) -> None:
        return self.selected_core_or_raise.write_memory_block32(addr, data)

    def read_memory_block8(self, addr: int, size: int) -> Sequence[int]:
        return self.selected_core_or_raise.read_memory_block8(addr, size)

    def read_memory_block32(self, addr: int, size: int) -> Sequence[int]:
        return self.selected_core_or_raise.read_memory_block32(addr, size)

    def read_core_register(self, id: CoreRegisterNameOrNumberType) -> CoreRegisterValueType:
        return self.selected_core_or_raise.read_core_register(id)

    def write_core_register(self, id: CoreRegisterNameOrNumberType, data: CoreRegisterValueType) -> None:
        return self.selected_core_or_raise.write_core_register(id, data)

    def read_core_register_raw(self, reg: CoreRegisterNameOrNumberType) -> int:
        return self.selected_core_or_raise.read_core_register_raw(reg)

    def read_core_registers_raw(self, reg_list: Sequence[CoreRegisterNameOrNumberType]) -> List[int]:
        return self.selected_core_or_raise.read_core_registers_raw(reg_list)

    def write_core_register_raw(self, reg: CoreRegisterNameOrNumberType, data: int) -> None:
        self.selected_core_or_raise.write_core_register_raw(reg, data)

    def write_core_registers_raw(self, reg_list: Sequence[CoreRegisterNameOrNumberType], data_list: Sequence[int]) -> None:
        self.selected_core_or_raise.write_core_registers_raw(reg_list, data_list)

    def find_breakpoint(self, addr: int) -> Optional[Breakpoint]:
        return self.selected_core_or_raise.find_breakpoint(addr)

    def set_breakpoint(self, addr: int, type: Target.BreakpointType = Target.BreakpointType.AUTO) -> bool:
        return self.selected_core_or_raise.set_breakpoint(addr, type)

    def get_breakpoint_type(self, addr: int) -> Optional[Target.BreakpointType]:
        return self.selected_core_or_raise.get_breakpoint_type(addr)

    def remove_breakpoint(self, addr: int) -> None:
        return self.selected_core_or_raise.remove_breakpoint(addr)

    def set_watchpoint(self, addr: int, size: int, type: Target.WatchpointType) -> bool:
        return self.selected_core_or_raise.set_watchpoint(addr, size, type)

    def remove_watchpoint(self, addr: int, size: Optional[int], type: Optional[Target.WatchpointType]) -> None:
        return self.selected_core_or_raise.remove_watchpoint(addr, size, type)

    def reset(self, reset_type: Optional[Target.ResetType] = None) -> None:
        # Use the probe to reset to perform a hardware reset if there is not a core.
        if self.selected_core is None:
            # Use the probe to reset. (We can't use the DP here because that's a class layering violation;
            # the DP is only created by the CoreSightTarget subclass.)
            assert self.session.probe
            self.session.probe.reset()
            return
        self.selected_core_or_raise.reset(reset_type)

    def reset_and_halt(self, reset_type: Optional[Target.ResetType] = None) -> None:
        return self.selected_core_or_raise.reset_and_halt(reset_type)

    def get_state(self) -> Target.State:
        return self.selected_core_or_raise.get_state()

    def get_security_state(self) -> Target.SecurityState:
        return self.selected_core_or_raise.get_security_state()

    def get_halt_reason(self) -> Target.HaltReason:
        return self.selected_core_or_raise.get_halt_reason()

    def set_vector_catch(self, enable_mask: int) -> None:
        return self.selected_core_or_raise.set_vector_catch(enable_mask)

    def get_vector_catch(self) -> int:
        return self.selected_core_or_raise.get_vector_catch()

    def get_target_context(self, core: Optional[int] = None) -> DebugContext:
        if core is not None:
            core_obj = self.cores[core]
        else:
            core_obj = self.selected_core_or_raise
        return core_obj.get_target_context()

    def trace_start(self):
        self.call_delegate('trace_start', target=self, mode=0)

    def trace_stop(self):
        self.call_delegate('trace_stop', target=self, mode=0)

    def add_target_command_groups(self, command_set: CommandSet):
        """@brief Hook for adding target-specific commands to a command set."""
        self.call_delegate('add_target_command_groups', target=self, command_set=command_set)




# pyOCD debugger
# Copyright (c) 2020 Bartek Wolowiec
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import logging
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...coresight.cortex_m import CortexM
from ...coresight.minimal_mem_ap import MinimalMemAP as MiniAP

LOG = logging.getLogger(__name__)

class DBGMCU:
    CR = 0xE0042004
    CR_VALUE = 0x7 # DBG_STANDBY | DBG_STOP | DBG_SLEEP

FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x8f4ff3bf, 0x02c04770, 0x28400dc0, 0x0980d302, 0x47701d00, 0xd3022820, 0x1cc00940, 0x08c04770,
    0x48494770, 0x60414947, 0x60414948, 0x60012100, 0x22f068c1, 0x60c14311, 0x06806940, 0x4845d406,
    0x60014943, 0x60412106, 0x60814943, 0x47702000, 0x6901483d, 0x43110542, 0x20006101, 0xb5104770,
    0x69014839, 0x43212404, 0x69016101, 0x431103a2, 0x493a6101, 0xe0004a37, 0x68c36011, 0xd4fb03db,
    0x43a16901, 0x20006101, 0xb530bd10, 0xffbbf7ff, 0x68ca492d, 0x431a23f0, 0x240260ca, 0x690a610c,
    0x0e0006c0, 0x610a4302, 0x03e26908, 0x61084310, 0x8f4ff3bf, 0x4a274829, 0x6010e000, 0x03ed68cd,
    0x6908d4fb, 0x610843a0, 0x060068c8, 0xd0030f00, 0x431868c8, 0x200160c8, 0xb5f0bd30, 0x1cc94c1a,
    0x68e50889, 0x23f00089, 0x60e5431d, 0x61232300, 0x06ff2701, 0xe0214d19, 0x4e196923, 0x61234333,
    0x0af602c6, 0x681319f6, 0xf3bf6033, 0x4e118f4f, 0x6035e000, 0x03db68e3, 0x6923d4fb, 0x005b085b,
    0x68e36123, 0x0f1b061b, 0x68e0d005, 0x430821f0, 0x200160e0, 0x1d00bdf0, 0x1d121f09, 0xd1db2900,
    0xbdf02000, 0x45670123, 0x40023c00, 0xcdef89ab, 0x00005555, 0x40003000, 0x00000fff, 0x0000aaaa,
    0x00000201, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000043,
    'pc_unInit': 0x20000071,
    'pc_program_page': 0x200000fb,
    'pc_erase_sector': 0x200000ab,
    'pc_eraseAll': 0x2000007f,

    'static_base' : 0x20000000 + 0x00000020 + 0x00000164,
    'begin_stack' : 0x20000400,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x400,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000, 0x20001400],   # Enable double buffering
    'min_program_length' : 0x400
    }

class STM32F767xx(CoreSightTarget):

    VENDOR = "STMicroelectronics"

    # MemoryMap for dual bank configuration (Details in AN4826).
    # Dual memory configuration is controled by nDBANK bit in FLASH_OPTCR.
    # In both configuration there is 2MB of flash memory. The difference is in
    # flash sectors structure.
    # For Single bank there is 12 sectors, for dual bank there is 24 sectors.
    # For dual bank configurations sectors are half size.
    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x08000000, length=0x20000,  sector_size=0x8000,
                                                        page_size=0x400,
                                                        is_boot_memory=True,
                                                        algo=FLASH_ALGO),

        FlashRegion( start=0x08020000, length=0x20000,  sector_size=0x20000,
                                                        page_size=0x400,
                                                        algo=FLASH_ALGO),

        FlashRegion( start=0x08040000, length=0x1C0000,  sector_size=0x40000,
                                                        page_size=0x400,
                                                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x80000)
        )

    def __init__(self, session):
        super(STM32F767xx, self).__init__(session, self.MEMORY_MAP)

    def assert_reset_for_connect(self):
        self.dp.assert_reset(1)

    def safe_reset_and_halt(self):
        assert self.dp.is_reset_asserted()

        # At this point we can't access full AP as it is not initialized yet.
        # Let's create a minimalistic AP and use it.
        ap = MiniAP(self.dp)
        ap.init()

        DEMCR_value = ap.read32(CortexM.DEMCR)

        # Halt on reset.
        ap.write32(CortexM.DEMCR, CortexM.DEMCR_VC_CORERESET)
        ap.write32(CortexM.DHCSR, CortexM.DBGKEY | CortexM.C_DEBUGEN)

        # Prevent disabling bus clock/power in low power modes.
        ap.write32(DBGMCU.CR, DBGMCU.CR_VALUE)

        self.dp.assert_reset(0)
        time.sleep(0.01)

        # Restore DEMCR original value.
        ap.write32(CortexM.DEMCR, DEMCR_value)

    def create_init_sequence(self):
        # STM32 under some low power/broken clock states doesn't allow AHP communication.
        # Low power modes are quite popular on stm32 (including MBed OS defaults).
        # 'attach' mode is broken by default, as STM32 can't be connected on low-power mode
        #  successfully without previous DBGMCU setup (It is not possible to write DBGMCU).
        # It is also not possible to run full pyOCD discovery code under-reset.
        #
        # As a solution we can setup DBGMCU under reset, halt core and release reset.
        # Unfortunately this code has to be executed _before_ discovery stage
        # and without discovery stage we don't have access to AP/Core.
        # As a solution we can create minimalistic AP implementation and use it
        # to setup core halt.
        # So the sequence for 'halt' connect mode will look like
        # -> Assert reset
        # -> Connect DebugPort
        # -> Setup MiniAp
        # -> Setup halt on reset
        # -> Enable support for debugging in low-power modes
        # -> Release reset
        # -> [Core is halted and reset is released]
        # -> Continue [discovery, create cores, etc]
        seq = super(STM32F767xx, self).create_init_sequence()
        if self.session.options.get('connect_mode') in ('halt', 'under-reset'):
            seq.insert_before('dp_init', ('assert_reset_for_connect', self.assert_reset_for_connect))
            seq.insert_after('dp_init', ('safe_reset_and_halt', self.safe_reset_and_halt))

        return seq

    def post_connect_hook(self):
        FLASH_OPTCR_ADDR = 0x40023C14
        FLASH_OPTCR_NDBANK = 1<<29
        flash_optcr = self.read32(FLASH_OPTCR_ADDR)
        if flash_optcr & FLASH_OPTCR_NDBANK:
            LOG.info('Single bank configuration detected [FLASH_OPTCR=0x%08x].', flash_optcr)
        else:
            LOG.error('Dual bank configuration detected [FLASH_OPTCR=0x%08x]. '
                      'Currently only single bank configuration is fully supported!', flash_optcr)
        self.write32(DBGMCU.CR, DBGMCU.CR_VALUE)


